//! \example _tests/fsConWin/main.cpp
//---------------------------------------------------------------------------
#include <fs/rap/fsConWin.h>

#include <cassert>
#include <string>
#include <windows.h>
//---------------------------------------------------------------------------
namespace test {
//---------------------------------------------------------------------------
//! A sample text receiver, it routes typed text to the target console.
class TxtRcvr_t : public fs::rap::ConWin_t::TxtRcvr_i
{
 public:
 
    explicit TxtRcvr_t(fs::rap::ConWin_t *pTgt) : m_pTgt(pTgt)
     { assert(m_pTgt); }

    virtual void Receive(const char *pszText, bool bCmd)
    {
        assert(pszText);
    
        if(!bCmd) // process only command-line inputs
            return;
            
        if(::strcmp(pszText, "exit") == 0) // quit the application
            ::PostQuitMessage(0);
        else
        {
            // send typed text to the target console
            const std::string strText("RECEIVED: " + std::string(pszText));
            m_pTgt->Push(strText.c_str());
        }
    }
            
 private:
 
    TxtRcvr_t();
    fs::rap::ConWin_t *m_pTgt;
};
//---------------------------------------------------------------------------
} // namespace test
//---------------------------------------------------------------------------
int WINAPI WinMain(HINSTANCE, HINSTANCE, LPSTR, int)
{
    // create two consoles
    fs::rap::ConWin_t Con1("Console 1");
    assert(Con1.IsValid());
    
    fs::rap::ConWin_t Con2("Console 2");
    assert(Con2.IsValid());
    
    // register text receivers (cross input <=> output)
    test::TxtRcvr_t Rcvr1(&Con2);
    Con1.RegTxtRcvr(&Rcvr1);
    
    test::TxtRcvr_t Rcvr2(&Con1);
    Con2.RegTxtRcvr(&Rcvr2);
    
    // set positions and dimensions of the consoles
    Con1.SetPos(20, 20,  750, 260);
    Con2.SetPos(20, 300, 750, 260);
    
    // show the consoles
    Con1.Show();
    Con2.Show();
    
    // send a sample message to the output of each of the consoles
    Con1.Push("Sample message 1. Type 'exit' to quit the application.");
    Con2.Push("Sample message 2. Type 'exit' to quit the application.");
            
    //--- MESSAGE PUMP ---
        
    // usually the underlying application has such a pump...
    MSG msg;
    while(::GetMessage(&msg, 0, 0, 0))
    {
        ::TranslateMessage(&msg);
        ::DispatchMessage(&msg);
    }

    return static_cast<int>(msg.wParam);
}
//---------------------------------------------------------------------------