//! \example _tests/fsLog/main.cpp
//---------------------------------------------------------------------------
#include <fs/sys/fsLog.h>

#include <fs/sys/fsDll.h>

#include <cassert>
//---------------------------------------------------------------------------
int main()
{
    using namespace fs::sys;

    FS_LOG("o.EXE The application has been launched.");    
    FS_LOG("!.EXE Sample warning message!");
    FS_LOG("x.EXE Sample error message!");
    FS_LOG("o.EXE Still logs only into the log history...");
        
    static Log_t AppLog(Log_t::LOG_FILE | Log_t::LOG_STDOUT);
    ModLog().SetTgt(&AppLog); // link 'module log' => 'app log'
    
    FS_LOG("o.EXE First routed (real-time) message.");
                
    FS_LOG("o.EXE At this point a module is loaded.");
    { 
        // load a module
        fs::sys::Dll_t Dll("fsLogDll.dll");
        assert(Dll.IsValid());
        
        // export a function from the module
        typedef void (* Bind_fn)(Log_t *, bool);
        Bind_fn pfnBind = static_cast<Bind_fn>(Dll.GetSymb("BindAppLog"));
        assert(pfnBind);
        
        // bind 'Dll (module) Log' => 'AppLog', set 'AppLog' as a target
        (*pfnBind)(&AppLog, true);
                        
        // export a function from the module
        typedef int (* SomeDllFunc_fn)(int);
        SomeDllFunc_fn pfnSomeDllFunc = static_cast<SomeDllFunc_fn>
         (Dll.GetSymb("SomeDllFunc"));
        assert(pfnSomeDllFunc);

        // call the function exported from the module        
        int nRes = (*pfnSomeDllFunc)(10);
        FS_LOG("o.EXE The result of SomeDllFunc(10) is : %d", nRes);
        
        // unbind 'Dll (module) Log', set no target
        (*pfnBind)(&AppLog, false);
    }
    FS_LOG("o.EXE At this point the module is unloaded.");
    
    FS_LOG("o.EXE The application is about to be terminated.");
           
    return 0;
}
//---------------------------------------------------------------------------
// [000000:015] o.EXE The application has been launched.
// [000000:015] !.EXE Sample warning message!
// [000000:015] x.EXE Sample error message!
// [000000:015] o.EXE Still logs only into the log history...
// [000000:015] o.EXE First routed (real-time) message.
// [000000:015] o.EXE At this point a module is loaded.
// [000000:015] o.DLL The module has just been attached.
// [000000:015] o.DLL Still logs only into the log history...
// [000000:015] o.DLL SomeDllFunc(10) has just been called.
// [000000:015] o.EXE The result of SomeDllFunc(10) is : 11
// [000000:015] o.EXE At this point the module is unloaded.
// [000000:015] o.EXE The application is about to be terminated.
// o.LOG 1 warning(s), 1 error(s) found.
//---------------------------------------------------------------------------