//! \example _tests/fsMemMgr/main.cpp
//---------------------------------------------------------------------------
// Please run this example from the development environment to see a proper
// code position in case a RUNTIME BREAK occurs.
//---------------------------------------------------------------------------
// Turn on the memory tracking & the wrapping macros (which provide detailed
// code position information). It should be included as the first header in
// a common configuration header.
#include <fs/sys/fsMemMgr.h>
//---------------------------------------------------------------------------
// Most of headers (those that doesn't collide with the 'new' and 'delete'
// wrapping macros can be included freely).
// ... here come headers mentioned above ...
#include <stdio.h>
//---------------------------------------------------------------------------
// If the headers collide, the wrapping 'new' and 'delete' macros have to be
// disabled temporarily (note that this technique in not limited to header
// including).
#include <fs/sys/fsMemMgr_NoT.h> // turn off the tracking macros
// ... here come guarded headers...
#include <vector>
#include <fs/sys/fsMemMgr.h> // turn on the tracking macros
//---------------------------------------------------------------------------
void SomeFunc()
{
    std::vector<int *> Leak3;
    Leak3.push_back(new int);
    Leak3.push_back(new int);
}
//---------------------------------------------------------------------------
int main()
{
    // optional - affects the name of the log file
    fs::sys::mmgr::SetModNm("main");

    int *pLeak1 = new int;
    char *pLeak2 = new char[128];
    
    SomeFunc();  
 
    // UNKNOWN / BAD MEMORY ADDRESS => MEMORY MANAGER RUNTIME BREAK
    // <uncomment below to see>
    // int *pBadPtr = pLeak1 + 1;
    // delete pBadPtr;
    
    // WRONG DEALLOCATION METHOD => MEMORY MANAGER RUNTIME BREAK
    // <uncomment below to see>
    // delete pLeak2;
    
    // ....
    // MEMORY LEAKS DETECTED => MEMORY MANAGER RUNTIME BREAK
    
    return 0;
}
//---------------------------------------------------------------------------
// +--------------------------------------------+
// | FipS' MEMORY MANAGER (basic stats)         |
// | module: main                               |
// +--------------------------------+-----------+
// |    Total number of allocations | 6         |
// |  Total number of deallocations | 2         |
// |      Max number of allocations | 6         |
// |                Number of leaks | 4       ! |
// +--------------------------------+-----------+

// +------------------------------------------+
// | FipS' MEMORY MANAGER (tracking details)  |
// | module: main                             |
// +------------------------------+-----------+
// |    Max allocated memory size | 152 B     |
// |              Number of leaks | 4       ! |
// +------------------------------+--+--------+---------------------------------+------------+-----------+
// | ...\fs\_tests\fsmemmgr\main.cpp | 36     | main                            | 0x00356180 | 4 B       |
// | ...\fs\_tests\fsmemmgr\main.cpp | 37     | main                            | 0x00356260 | 128 B     |
// | ...\fs\_tests\fsmemmgr\main.cpp | 27     | SomeFunc                        | 0x003563c0 | 4 B       |
// | ...\fs\_tests\fsmemmgr\main.cpp | 28     | SomeFunc                        | 0x00356580 | 4 B       |
// +---------------------------------+--------+---------------------------------+------------+-----------+
