//! \example _tests/fsRle/main.cpp
//---------------------------------------------------------------------------
#include <fs/ras/fsRle8Enc.h>
#include <fs/ras/fsRle8Blt.h>

#include <stdio.h> // printf
#include <memory.h> // memset, memcmp
#include <cassert>
//---------------------------------------------------------------------------
//! Helper for conversion between compatible pixel formats.
const fs::ras::nsRle8Blt::PxFmt8888_t *
Conv(const fs::ras::Rle8Enc_t::PxFmt8888_t *pFrom)
{
    // put a static assert here to ensure the types are equal in size
    return reinterpret_cast<const fs::ras::nsRle8Blt::PxFmt8888_t *>(pFrom);
}
//---------------------------------------------------------------------------
int main()
{
    using namespace fs::ras;

    // input RGBA-8888 image data
    const Rle8Enc_t::PxFmt8888_t aIn[] = {
     {0,0,0,0},       {0,0,255,128},   {0,0,0,0},
     {0,255,0,128},   {255,0,0,255},   {0,255,0,128},
     {0,0,255,128},   {0,0,255,128},   {0,0,0,0},
    };
    
    // encode RGBA-8888 image data into RLE8 data stream
    Rle8Enc_t Rle(aIn, 3, 3);
    assert(Rle.IsValid() && "Couldn't encode image data!");
        
    // get image dimensions
    int nW = 0, nH = 0;
    Rle.GetSize(nW, nH);
    ::printf("image width = %d, image height = %d\n", nW, nH);
        
    // get RLE8 data stream
    const unsigned char *pRle = 0;
    int nRleNum = 0;
    Rle.GetRle(pRle, nRleNum);
    ::printf("RLE8 data stream: ");
    for(int i = 0; i < nRleNum; ++i)
    {
        bool bTr = (pRle[i] & 1) == 0;
        int nNum = pRle[i] >> 1;

        ::printf("%d{%d/%d} ", pRle[i], !bTr, nNum);

        if(!bTr)
            for(int j = 0; j < nNum; ++i, ++j)
                ::printf("%d ", pRle[i + 1]);

    }
    ::printf("\n");
    
    // get palette items
    const Rle8Enc_t::PxFmt8888_t *pPal = 0;
    int nPalNum = 0;
    Rle.GetPal(pPal, nPalNum);
    ::printf("palette items: ");
    for(int i = 0; i < nPalNum; ++i)
    {
        ::printf("%d-%d-%d-%d ",
         pPal[i].r, pPal[i].g, pPal[i].b, pPal[i].a);
    }
    ::printf("\n");
    
    // get line offsets
    const Rle8Enc_t::uint32 *pOff = 0;
    int nOffNum = 0;
    Rle.GetOff(pOff, nOffNum);
    ::printf("line offsets: ");
    for(int i = 0; i < nOffNum; ++i)
        ::printf("%d ", pOff[i]);
    ::printf("\n---\n");

    // decode RLE8 data stream back to RGBA-8888 image data
    nsRle8Blt::PxFmt8888_t aOut[sizeof(aIn)];
    ::memset(aOut, 0, sizeof(aOut));
    
    Rle8Blt<nsRle8Blt::Cl8888TrCpPcy_t>(
     pRle, Conv(pPal), pOff,
     3, 3,
     3, 3,
     aOut,
     3, 3,
     0, 0
    );
    
    // compare the input with the output
    ::printf("aIn %s aOut\n", ::memcmp(aIn, aOut, sizeof(aIn)) == 0 ?
     "==" : "!=");
    
    return 0;
}
//---------------------------------------------------------------------------
// image width = 3, image height = 3
// RLE8 data stream: 2{0/1} 3{1/1} 0 2{0/1} 7{1/3} 1 2 1 5{1/2} 0 0 2{0/1}
// palette items: 0-0-255-128 0-255-0-128 255-0-0-255
// line offsets: 0 4 8
// ---
// aIn == aOut
//---------------------------------------------------------------------------