//! \example _tests/fsStrUt/main.cpp
//---------------------------------------------------------------------------
#include <fs/ut/fsStrUt.h>

#include <cstdio> // printf
#include <cstring> // strlen
//---------------------------------------------------------------------------
int main()
{
    using namespace fs::ut::str;
            
    //--- MISC STRING OPERATIONS ---
    
    ::printf("MISC STRING OPERATIONS:\n");
    
    StrHldr_t Str1("String 1 - copied");
    ::printf("%s\n", Str1.GetStr());
    
    StrHldr_t Str2(L"String 2 - converted from a wide-character string");
    ::printf("%s\n", Str2.GetStr());
    
    WStrHldr_t WStr1(L"Wide String 1 - copied");
    ::printf("%S\n", WStr1.GetWStr());

    WStrHldr_t WStr2("Wide String 2 - converted from a single-byte string");
    ::printf("%S\n", WStr2.GetWStr());
    
    //--- STRING TOKENIZER ---
    
    const char pszString[] = "one, two, \"t h r e e\"";
    
    ::printf("STRING TOKENIZER: %s\n", pszString);
            
    Toknzr_t Toknzr(pszString, " ,");
    for(const char *pszTok = Toknzr.Next(); pszTok; pszTok = Toknzr.Next())
    {
        ::printf("token: %s\n", pszTok);
    }

    //--- COMMAND LINE PARSER ---
    
    ::printf("COMMAND LINE PARSER\n");
    
    CmdLn_t CmdLn("foo -key1 value1 -switch -key2 \"complex value2\"");
    
    for(int i = 0, n = CmdLn.GetNumToks(); i < n; ++i)
        ::printf("token[%d] : %s\n", i, CmdLn.GetTok(i));

    const char *pszVal1 = CmdLn.GetVal("key1");
    ::printf("key1: %s\n", pszVal1 ? pszVal1 : "<not found>");
    const char *pszVal2 = CmdLn.GetVal("key2");
    ::printf("key2: %s\n", pszVal2 ? pszVal2 : "<not found>");
    const char *pszVal3 = CmdLn.GetVal("key3");
    ::printf("key3: %s\n", pszVal3 ? pszVal3 : "<not found>");
    const bool bSwitch = CmdLn.FindTok("-switch") >= 0;
    ::printf("-switch: %s\n", bSwitch ? "true" : "false");
                
    //---
    
    return 0;
}
//---------------------------------------------------------------------------
// MISC STRING OPERATIONS:
// String 1 - copied
// String 2 - converted from a wide-character string
// Wide String 1 - copied
// Wide String 2 - converted from a single-byte string
// STRING TOKENIZER: one, two, "t h r e e"
// token: one
// token: two
// token: t h r e e
// COMMAND LINE PARSER
// token[0] : foo
// token[1] : -key1
// token[2] : value1
// token[3] : -switch
// token[4] : -key2
// token[5] : complex value2
// key1: value1
// key2: complex value2
// key3: <not found>
// -switch: true
//---------------------------------------------------------------------------