//! \example _tests/fsTgaLdr/main.cpp
//---------------------------------------------------------------------------
#include <fs/ras/fsTgaLdr.h>

#include <stdio.h> // printf
#include <memory.h> // memset, memcmp
#include <cassert>
//---------------------------------------------------------------------------
int main()
{
    using namespace fs::ras;
        
    //--- TEST1 RGB-888 ---
    
    TgaLdr_t Tga1("test_rgb.tga");
    assert(Tga1.IsValid() && "Couldn't load the TGA file!");

    ::printf("---> IMAGE 1:\n");
    ::printf("width: %d\n", Tga1.GetW());
    ::printf("height: %d\n", Tga1.GetH());
    ::printf("is RGB-888 : %s\n", Tga1.GetData888() ? "true" : "false");
    ::printf("is RGBA-8888 : %s\n", Tga1.GetData8888() ? "true" : "false");
    
    // expected RGB-888 data
    const unsigned char aExpRgbData[] = {
     255, 255, 255,   255, 0, 0,   255, 255, 255,
     255, 255, 255,   0, 255, 0,   255, 255, 255,
     0,   0,   0,     0, 0, 255,   0,   0,   0,
    };

    // compare loaded and expected RGB-888 data
    assert(Tga1.GetData888());
    ::printf("aExpRgbData %s TGA Data\n", ::memcmp(Tga1.GetData888(),
     aExpRgbData, sizeof(aExpRgbData)) == 0 ? "==" : "!=");
    
    //--- TEST2 RGBA-8888 ---
    
    TgaLdr_t Tga2("test_rgba.tga");
    assert(Tga2.IsValid() && "Couldn't load the TGA file!");
        
    ::printf("---> IMAGE 2:\n");
    ::printf("width: %d\n", Tga2.GetW());
    ::printf("height: %d\n", Tga2.GetH());
    ::printf("is RGB-888 : %s\n", Tga2.GetData888() ? "true" : "false");
    ::printf("is RGBA-8888 : %s\n", Tga2.GetData8888() ? "true" : "false");
                
    // expected RGBA-8888 data
    const unsigned char aExpRgbaData [] = {
     255, 255, 255, 63,   255, 0, 0, 255,   255, 255, 255, 127,
     255, 255, 255, 63,   0, 255, 0, 255,   255, 255, 255, 127,
     0,   0,   0,   63,   0, 0, 255, 255,   0,   0,   0,   127,
    };
                    
    // compare loaded and expected RGBA-8888 data
    assert(Tga2.GetData8888());
    ::printf("aExpRgbaData %s TGA Data\n---\n", ::memcmp(Tga2.GetData8888(),
     aExpRgbaData, sizeof(aExpRgbaData)) == 0 ? "==" : "!=");
     
    //---

    return 0;
}
//---------------------------------------------------------------------------
// ---> IMAGE 1:
// width: 3
// height: 3
// is RGB-888 : true
// is RGBA-8888 : false
// aExpRgbData == TGA Data
// ---> IMAGE 2:
// width: 3
// height: 3
// is RGB-888 : false
// is RGBA-8888 : true
// aExpRgbaData == TGA Data
//---------------------------------------------------------------------------