/*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*/
/*                          _______      ______    ______          __      */
/*  ~ ~ ~ ~ ~ ~ ~ ~ ~ ~    / ____(_)___ / ___ /)  / ____/___  ____/ /__    */
/*      [fsFunc]          / /_  / / __ \\__ \|/  / /   / __ \/ __  / _ \   */
/*       rev. 2          / __/ / / /_/ /__/ /   / /___/ /_/ / /_/ /  __/   */
/*    20th Jan 2007     /_/   /_/ ,___/____/    \____/\____/\__,_/\___/    */
/*     [x] stable              /_/ (c) 2006-7 Filip STOKLAS (FipS)         */
/*  ~ ~ ~ ~ ~ ~ ~ ~ ~ ~       http://HOLE.4FipS.com/fips_code.php          */
/*                                                                         */
/* This code is free for personal and commercial use. You may redistribute */
/* it by any means. If you use the code for your own projects please give  */
/* me credit. Please send a bug report. Don't alter or remove this header! */
/*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*/
#ifndef FS_MTH_FSFUNC_H
#define FS_MTH_FSFUNC_H
//---------------------------------------------------------------------------
#if defined(FS_INCLUDE_USERDEFS) // inject user definition file on request
#   include <fs/fsUserDefs.h>
#endif

#include "fsConst.h"

#include <cmath>
//---------------------------------------------------------------------------
namespace fs { namespace mth {
//---------------------------------------------------------------------------
template<typename T>
inline T Min(T tVal1, T tVal2)
{
    return tVal1 < tVal2 ? tVal1 : tVal2;
}
//---------------------------------------------------------------------------
template<typename T>
inline T Max(T tVal1, T tVal2)
{
    return tVal1 > tVal2 ? tVal1 : tVal2;
}
//---------------------------------------------------------------------------
template<typename T>
inline T Clamp(T tVal, T tMinVal, T tMaxVal)
{
    return tVal < tMinVal ? tMinVal : tVal > tMaxVal ? tMaxVal : tVal;
}
//---------------------------------------------------------------------------
template<typename T>
inline T Abs(T tVal)
{
    return tVal >= static_cast<T>(0) ? tVal : -tVal;
}
//---------------------------------------------------------------------------
template <typename T>
inline T Sgn(T tVal)
{
    return tVal > static_cast<T>(0) ? static_cast<T>(1) :
     (tVal < static_cast<T>(0) ? static_cast<T>(-1) : static_cast<T>(0));
}
//---------------------------------------------------------------------------
template <typename T>
inline T Floor(T tVal)
{
    return static_cast<T>(::floor(tVal));
}
//---------------------------------------------------------------------------
inline float Sin(float fAngle)
{
    return ::sin(fAngle);
}
//---------------------------------------------------------------------------
inline float Cos(float fAngle)
{
    return ::cos(fAngle);
}
//---------------------------------------------------------------------------
inline void SinCos(float tAngle, float &tSin, float &tCos)
{
    tSin = ::sin(tAngle);
    tCos = ::cos(tAngle);
}
//---------------------------------------------------------------------------
template<typename T>
inline T Sqrt(T tVal)
{
    return static_cast<T>(::sqrt(tVal));
}
//---------------------------------------------------------------------------
template<typename T>
inline T SqrtInv(T tVal)
{
    return static_cast<T>(1) / Sqrt(tVal);
}
//---------------------------------------------------------------------------
template<typename T>
inline T Norm(T tVal, T tMin, T tMax)
{
    T tRange = tMax - tMin;
    return tVal - tRange * Floor((tVal - tMin) / tRange);
}
//---------------------------------------------------------------------------
inline int Norm0_360(int nVal)
{
    return (nVal % 360 + 360) % 360;
}
//---------------------------------------------------------------------------
inline float Norm0_360(float fVal)
{
    return fVal - 360.f * ::floor(fVal / 360.f);
}
//---------------------------------------------------------------------------
}} // namespace fs::mth
//---------------------------------------------------------------------------
#endif // FS_MTH_FSFIXED_H