/*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*/
/*                          _______      ______    ______          __      */
/*  ~ ~ ~ ~ ~ ~ ~ ~ ~ ~    / ____(_)___ / ___ /)  / ____/___  ____/ /__    */
/*      [fsVec2]          / /_  / / __ \\__ \|/  / /   / __ \/ __  / _ \   */
/*       rev. 3          / __/ / / /_/ /__/ /   / /___/ /_/ / /_/ /  __/   */
/*    20th Jan 2007     /_/   /_/ ,___/____/    \____/\____/\__,_/\___/    */
/*     [x] stable              /_/ (c) 2006-7 Filip STOKLAS (FipS)         */
/*  ~ ~ ~ ~ ~ ~ ~ ~ ~ ~       http://HOLE.4FipS.com/fips_code.php          */
/*                                                                         */
/* This code is free for personal and commercial use. You may redistribute */
/* it by any means. If you use the code for your own projects please give  */
/* me credit. Please send a bug report. Don't alter or remove this header! */
/*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*/
#ifndef FS_MTH_FSVEC2_H
#define FS_MTH_FSVEC2_H
//---------------------------------------------------------------------------
#if defined(FS_INCLUDE_USERDEFS) // inject user definition file on request
#   include <fs/fsUserDefs.h>
#endif
//---------------------------------------------------------------------------
namespace fs { namespace mth {
//---------------------------------------------------------------------------
//! Two dimensional vector representation (incomplete).
template <typename T>
class Vec2_T
{
 public:
 
    // --- CONSTRUCTION ---
    
    Vec2_T();
    Vec2_T(T x, T y);
    Vec2_T(const Vec2_T &RHS);
    
    // --- ASSIGMENT ---
    
    Vec2_T & operator = (const Vec2_T &RHS);
    
    // ---
    
    bool operator == (const Vec2_T &RHS) const;
    bool operator != (const Vec2_T &RHS) const;
                 
    // --- OPERATIONS ---
    
    Vec2_T operator - () const;
    Vec2_T operator + (const Vec2_T &RHS) const;
    Vec2_T & operator += (const Vec2_T &RHS);
    Vec2_T operator - (const Vec2_T &RHS) const;
    Vec2_T & operator -= (const Vec2_T &RHS);
    T operator * (const Vec2_T &RHS) const; // dot
    Vec2_T operator * (T t) const;
    // friend Vec2_T operator * (T t, const Vec2_T &RHS);
    Vec2_T & operator *= (T t);
    
    //! Returns the length.
    T Len() const;
    //! Returns the squared length.
    T Len2() const;
    //! Normalizes the vector and return its original length.
    T Norm();
 
 public: // direct access is allowed
 
    T x, y;
};
//---------------------------------------------------------------------------
template <typename T>
inline Vec2_T<T>::Vec2_T() // does nothing
{
}
//---------------------------------------------------------------------------
template <typename T>
inline Vec2_T<T>::Vec2_T(T x, T y):
x(x),
y(y)
{
}
//---------------------------------------------------------------------------
template <typename T>
inline Vec2_T<T>::Vec2_T(const Vec2_T &RHS):
x(RHS.x),
y(RHS.y)
{
}
//---------------------------------------------------------------------------
template <typename T>
inline Vec2_T<T> & Vec2_T<T>::operator = (const Vec2_T &RHS)
{
    x = RHS.x;
    y = RHS.y;
    return *this;
}
//---------------------------------------------------------------------------
template <typename T>
inline bool Vec2_T<T>::operator == (const Vec2_T &RHS) const
{
    return x == RHS.x && y == RHS.y;
}
//---------------------------------------------------------------------------
template <typename T>
inline bool Vec2_T<T>::operator != (const Vec2_T &RHS) const
{
    return x != RHS.x || y != RHS.y;
}
//---------------------------------------------------------------------------
template <typename T>
inline Vec2_T<T> Vec2_T<T>::operator - () const
{
    return Vec2_T(-x, -y);
}
//---------------------------------------------------------------------------
template <typename T>
inline Vec2_T<T> Vec2_T<T>::operator + (const Vec2_T &RHS) const
{
    return Vec2_T(x + RHS.x, y + RHS.y);
}
//---------------------------------------------------------------------------
template <typename T>
inline Vec2_T<T> & Vec2_T<T>::operator += (const Vec2_T &RHS)
{
    x += RHS.x;
    y += RHS.y;
    return *this;
}
//---------------------------------------------------------------------------
template <typename T>
inline Vec2_T<T> Vec2_T<T>::operator - (const Vec2_T &RHS) const
{
    return Vec2_T(x - RHS.x, y - RHS.y);
}
//---------------------------------------------------------------------------
template <typename T>
inline Vec2_T<T> & Vec2_T<T>::operator -= (const Vec2_T &RHS)
{
    x -= RHS.x;
    y -= RHS.y;
    return *this;
}
//---------------------------------------------------------------------------
template <typename T>
inline T Vec2_T<T>::operator * (const Vec2_T &RHS) const
{
    return x * RHS.x + y * RHS.y;
}
//---------------------------------------------------------------------------
template <typename T>
inline Vec2_T<T> Vec2_T<T>::operator * (T t) const
{
    return Vec2_T(t * x, t * y);
}
//---------------------------------------------------------------------------
template <typename T>
inline Vec2_T<T> operator * (T t, const Vec2_T<T> &RHS)
{
    return Vec2_T<T>(t * RHS.x, t * RHS.y);
}
//---------------------------------------------------------------------------
template <typename T>
inline Vec2_T<T> & Vec2_T<T>::operator *= (T t)
{
    x *= t;
    y *= t;
    return *this;
}
//---------------------------------------------------------------------------
template <typename T>
inline T Vec2_T<T>::Len() const
{
    return Sqrt(Len2());
}
//---------------------------------------------------------------------------
template <typename T>
inline T Vec2_T<T>::Len2() const
{
    return (*this) * (*this);
}
//---------------------------------------------------------------------------
template <typename T>
inline T Vec2_T<T>::Norm()
{
    T l2 = Len2();
    T li = SqrtInv(l2);
    
    x *= li;
    y *= li;
    
    return li * l2; // returns length
}
//---------------------------------------------------------------------------
}} // namespace fs::mth
//---------------------------------------------------------------------------
#endif // FS_MTH_FSVEC2_H