/*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*/
/*                          _______      ______    ______          __      */
/*  ~ ~ ~ ~ ~ ~ ~ ~ ~ ~    / ____(_)___ / ___ /)  / ____/___  ____/ /__    */
/*     [fsConWin]         / /_  / / __ \\__ \|/  / /   / __ \/ __  / _ \   */
/*       rev. 4          / __/ / / /_/ /__/ /   / /___/ /_/ / /_/ /  __/   */
/*    8th Feb 2007      /_/   /_/ ,___/____/    \____/\____/\__,_/\___/    */
/*     [x] stable              /_/ (c) 2006-7 Filip STOKLAS (FipS)         */
/*  ~ ~ ~ ~ ~ ~ ~ ~ ~ ~       http://HOLE.4FipS.com/fips_code.php          */
/*                                                                         */
/* This code is free for personal and commercial use. You may redistribute */
/* it by any means. If you use the code for your own projects please give  */
/* me credit. Please send a bug report. Don't alter or remove this header! */
/*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*/
#ifndef FS_RAP_FSCONWIN_H
#define FS_RAP_FSCONWIN_H
//---------------------------------------------------------------------------
#if defined(FS_INCLUDE_USERDEFS) // inject user definition file on request
#   include <fs/fsUserDefs.h>
#endif 
//---------------------------------------------------------------------------
namespace fs { namespace rap {
//---------------------------------------------------------------------------
//! Console window.
/*! The class represents a console window. It consists of a <b>text output
    area</b> and a <b>command-line</b>. Text can be passed to the console
    using #Push or <b>a user can type</b> text directly to the command-line.
    The console then <b>notifies</b> registered text <b>receivers</b> (see
    #RegTxtRcvr) of such an activity.  Each instance of the class represents
    a single console window.
    
    \image html fs_rap_ConWin_t_01.gif
    
    Currently, <b>Windows</b> platform is the only supported. Console window
    requires a <b>message pump</b> to work properly. <b>GUI applications</b>
    typically have such a message pump so that's not the issue. In the case
    of <b>non-GUI applications</b> (e.g. DLLs) the <b>pump has to be
    provided</b> by the user e.g. in a separated thread. See a pump below
    for the basic idea.
    
    \code
    MSG msg;
    while(::GetMessage(&msg, 0, 0, 0))
    {
        ::TranslateMessage(&msg);
        ::DispatchMessage(&msg);
    }
    \endcode
*/
class ConWin_t
{
 public:
 
    struct TxtRcvr_i;
 
    //! Creates a new console window, which is by default invisible.
    explicit ConWin_t(const char *pszTitle = "FipS' Console");
    //! Destroys the console window.
    ~ConWin_t();
    //! Checks the internal state / correct construction.
    bool IsValid() const;
    
    //! Shows / hides the console window.
    void Show(bool bShow = true);
    //! Sets the position and dimensions of the console window.
    void SetPos(int nX, int nY, int nW, int nH);
    //! Sends a text line to the console output.
    void Push(const char *pszText);
    
    //! Registers a new text receiver.
    void RegTxtRcvr(TxtRcvr_i *pRcvr);
                    
 private:

    ConWin_t(const ConWin_t &); //!< No copy.
    ConWin_t & operator = (const ConWin_t &); //!< No assignment.

    void *m_pImpl; //!< The PIPML idiom is used.
    
 public:
 
    //! Text receiver interface. 
    struct TxtRcvr_i
    {
        virtual void Receive(const char *pszText, bool bCmd) = 0;
    };
};
//---------------------------------------------------------------------------
}} // namespace fs::rap
//---------------------------------------------------------------------------
#endif // FS_RAP_FSCONWIN_H