/*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*/
/*                          _______      ______    ______          __      */
/*  ~ ~ ~ ~ ~ ~ ~ ~ ~ ~    / ____(_)___ / ___ /)  / ____/___  ____/ /__    */
/*     [fsSurfCl]         / /_  / / __ \\__ \|/  / /   / __ \/ __  / _ \   */
/*       rev. 4          / __/ / / /_/ /__/ /   / /___/ /_/ / /_/ /  __/   */
/*    12th Feb 2007     /_/   /_/ ,___/____/    \____/\____/\__,_/\___/    */
/*     [ ] stable              /_/ (c) 2006-7 Filip STOKLAS (FipS)         */
/*  ~ ~ ~ ~ ~ ~ ~ ~ ~ ~       http://HOLE.4FipS.com/fips_code.php          */
/*                                                                         */
/* This code is free for personal and commercial use. You may redistribute */
/* it by any means. If you use the code for your own projects please give  */
/* me credit. Please send a bug report. Don't alter or remove this header! */
/*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*/
#ifndef FS_RAP_FSSURFCL_H
#define FS_RAP_FSSURFCL_H
//---------------------------------------------------------------------------
#if defined(FS_INCLUDE_USERDEFS) // inject user definition file on request
#   include <fs/fsUserDefs.h>
#endif

#include "fsWinApp.h" // WinApp_t::Event_t
//---------------------------------------------------------------------------
namespace fs { namespace rap {
//---------------------------------------------------------------------------
//! WinApp_t nested bitmap surface client (incomplete).
/*! The class represents a <b>bitmap surface</b>. It is typically used as a
    target for all sorts of bitmap and drawing operations. The class is
    closely related to WinApp_t. It implements WinApp_t::EvRcvr_i
    <b>interface</b> so it can be registered as an <b>event receiver</b>
    of WinApp_t, it acts as its <b>nested client</b>.
    
    \image html fs_rap_SurfCl_t_01.png
*/
class SurfCl_t : public WinApp_t::EvRcvr_i
{
 public:
    
    //! Surface pixel format.
    struct PxFmt8888_t { unsigned char r, g, b, x_; };
 
    //! Creates surface data using the given parameters.
    explicit SurfCl_t(bool bFitCl = false, int nSurfW = 512,
     int nSurfH = 512);
    //! Destroys surface data.
    virtual ~SurfCl_t();

    //! Clears the surface with the given color.
    void Clear(const PxFmt8888_t &Clr);
    //! Returns the width of the surface.
    int GetW() const;
    //! Returns the height of the surface.
    int GetH() const;
    //! Returns the surface data.
    PxFmt8888_t * GetData();
    
    virtual void OnEvent(const WinApp_t::SizeEv_t &, WinApp_t &);
    virtual void OnEvent(const WinApp_t::DrawEv_t &, WinApp_t &);
            
 private:

    SurfCl_t(const SurfCl_t &); //!< No copy.
    SurfCl_t & operator = (const SurfCl_t &); //!< No assignment.

    void *m_pImpl; //!< The PIPML idiom is used.
};
//---------------------------------------------------------------------------
}} // namespace fs::rap
//---------------------------------------------------------------------------
#endif // FS_RAP_FSSURFCL_H