/*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*/
/*                          _______      ______    ______          __      */
/*  ~ ~ ~ ~ ~ ~ ~ ~ ~ ~    / ____(_)___ / ___ /)  / ____/___  ____/ /__    */
/*     [fsBmfFnt]         / /_  / / __ \\__ \|/  / /   / __ \/ __  / _ \   */
/*       rev. 3          / __/ / / /_/ /__/ /   / /___/ /_/ / /_/ /  __/   */
/*    14th Feb 2007     /_/   /_/ ,___/____/    \____/\____/\__,_/\___/    */
/*     [ ] stable              /_/ (c) 2006-7 Filip STOKLAS (FipS)         */
/*  ~ ~ ~ ~ ~ ~ ~ ~ ~ ~       http://HOLE.4FipS.com/fips_code.php          */
/*                                                                         */
/* This code is free for personal and commercial use. You may redistribute */
/* it by any means. If you use the code for your own projects please give  */
/* me credit. Please send a bug report. Don't alter or remove this header! */
/*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*/
#include <fs/ras/fsBmfFnt.h>

// >>> COMMON SUPPORT BLOCK (rev. 6)
#if defined(FS_HAS_FSASSERT)
#   include <fs/sys/fsAssert.h>
#else
#   include <cassert>
#   if !defined(FS_ASSERT)
#       define FS_ASSERT(exp) assert(exp)
#   endif
#   if !defined(FS_ASSERT_MSG)
#       define FS_ASSERT_MSG(exp, msg) assert(exp && msg)
#   endif
#   if !defined(FS_VERIFY)
#       define FS_VERIFY(exp) \
        { bool bExp = !!(exp); assert(bExp && #exp); bExp; }
#   endif
#   if !defined(FS_VERIFY_MSG)
#       define FS_VERIFY_MSG(exp, msg) \
        { bool bExp = !!(exp); assert(bExp && #exp && msg); bExp; }
#   endif
#   if !defined(FS_STATIC_ASSERT)
#       define FS_STATIC_ASSERT(exp) { char error[(exp) ? 1 : 0]; error; }
#   endif
#endif
#if !defined(FS_VERIFY_RETURN)
#   define FS_VERIFY_RETURN(exp) \
    if(!(exp)) { FS_ASSERT_MSG(0, #exp "<return>"); return; }
#endif
#if !defined(FS_VERIFY_RETURN_VAL)
#   define FS_VERIFY_RETURN_VAL(exp, ret) \
    if(!(exp)) { FS_ASSERT_MSG(0, #exp "<returns:>" #ret); return (ret); }
#endif
// <<< COMMON SUPPORT BLOCK (rev. 6)

#include <vector>

// provide detailed memory tracking information
#if defined(FS_HAS_FSMEMMGR)
#   include <fs/sys/fsMemMgr.h>
#endif
//---------------------------------------------------------------------------
#if (_MSC_VER >= 1400) // MSVC8
#   pragma warning(disable : 4996) // function was marked as deprecated
#endif
//---------------------------------------------------------------------------
using namespace fs::ras;
//---------------------------------------------------------------------------
//--- ANONYMOUS NAMESPACE >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
//---------------------------------------------------------------------------
namespace {
//---------------------------------------------------------------------------
//--- CLASS BmfFntImpl_t >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
//---------------------------------------------------------------------------
class BmfFntImpl_t
{
 public:
            
    explicit BmfFntImpl_t(const char *pszFile);
    bool IsValid() const;
    
    void Print(int nX, int nY, const char *pszText, BmfFnt_t::PxFmt8888_t
     *pDst, int nDstW, int nDstH, int nClipX1, int nClipY1, int nClipX2,
     int nClipY2) const;

 private:
 
    BmfFntImpl_t(); //!< No def ctor.
    BmfFntImpl_t(const BmfFntImpl_t &); //!< No copy.
    BmfFntImpl_t & operator = (const BmfFntImpl_t &); //!< No assignment.
    
    //! BMF file header.
    struct Hdr_t
    {
        typedef signed char int8;
        typedef unsigned char uint8;
        typedef unsigned int uint32;
        
        bool m_bValid;
        uint32 m_uMagic; // E1 E6 D5 1A
        uint8 m_uVer; // 11h
        uint8 m_uLnH;
        int8 m_nOverLn;
        int8 m_nUnderLn;
        int8 m_nCharSpace;
        int8 m_nInnerSize;
        uint8 m_uNumClrs; // should be <= 32
        uint8 m_uMaxClr;
        uint32 m_uRsrvd;
        uint8 m_uNumRgb;
        const uint8 *m_pPal;
        uint8 m_uInfoLen;
        const uint8 *m_pInfo;
        uint8 m_uNumChars;
        const uint8 *m_pChars;

        Hdr_t() : m_bValid(false) {}
        explicit Hdr_t(const uint8 *pData);
    };
    
    //! BMF character representation.
    struct Char_t
    {
        typedef signed char int8;
        typedef unsigned char uint8;
    
        bool m_bValid;
        uint8 m_uID;
        uint8 m_uW;
        uint8 m_uH;
        int8 m_nRelX;
        int8 m_nRelY;
        uint8 m_uShift;
        const uint8 *m_pData;

        Char_t() : m_bValid(false) {}
        explicit Char_t(const uint8 *pData);
    };
    
    typedef std::vector<unsigned char> Data_t;
    typedef std::vector<Char_t> Chars_t;
        
    bool LoadData(const char *pszFile, Data_t &Data) const;
    void BindChars();
     
    //---
     
    Data_t m_Data;
    Hdr_t m_Hdr;
    Chars_t m_Chars;
};
//---------------------------------------------------------------------------
BmfFntImpl_t::BmfFntImpl_t(const char *pszFile)
{
    // --- LOAD BMF DATA (the whole content of the file) ---

    Data_t Data;
    
    if(!LoadData(pszFile, Data))
        return;
        
    //--- CHECK THE FILE TYPE ---
        
    {
        FS_ASSERT(!Data.empty());
        Hdr_t DummyHdr(&Data[0]);
    
        if(DummyHdr.m_uMagic != 0x1AD5E6E1)
            { FS_ASSERT_MSG(0, "Wrong BMF identifier!"); return; }

        if(DummyHdr.m_uVer !=  0x11)
            { FS_ASSERT_MSG(0, "Wrong BMF version!"); return; }
    }
        
    //--- TRANSFER THE STATE ---

    // swap the data
    m_Data.swap(Data);
    // bind the header (there are memory references to the data!)
    m_Hdr = Hdr_t(&m_Data[0]);
    // bind characters (there are memory references to the data!)
    BindChars();
}
//---------------------------------------------------------------------------
bool BmfFntImpl_t::IsValid() const
{
    return !m_Data.empty();
}
//---------------------------------------------------------------------------
void BmfFntImpl_t::Print(int nX, int nY, const char *pszText,
 BmfFnt_t::PxFmt8888_t *pDst,int nDstW, int /*nDstH*/, int /*nClipX1*/,
 int /*nClipY1*/, int /*nClipX2*/, int /*nClipY2*/) const
{
    FS_VERIFY_RETURN(pszText && pDst);
    
    for(int i = 0, nCurX = 0, nCurY = 0, n = static_cast<int>(::strlen
     (pszText)); i < n; ++i)
    {
        const char ch = pszText[i];
        const Char_t &Char = m_Chars[ch];
        
        if(!Char.m_bValid)
            continue; // skip undefined characters
            
        if(ch == '\n')
        {
            nCurX = 0;
            nCurY += m_Hdr.m_uLnH;
            continue;       
        }
            
        const unsigned char *pCharData = Char.m_pData;
        FS_ASSERT(pCharData);
        
        const int nSrcRelX = Char.m_nRelX;
        const int nSrcRelY = Char.m_nRelY;
        const int nSrcW = Char.m_uW;
        const int nSrcH = Char.m_uH;
        const int nDstBar = nSrcRelX + nX + nCurX;
        
        for(int y = 0; y < nSrcH; ++y)
        {
            const int nDstBlock = (y + nSrcRelY + nY + nCurY) * nDstW;
            
            const unsigned char *pCurSrc = pCharData + y * nSrcW;
            BmfFnt_t::PxFmt8888_t *pCurDst = pDst + nDstBlock + nDstBar;
                
            for(int x = 0; x < nSrcW; ++x)
            {
                int nColor = 3 * (*pCurSrc - 1);
                if(nColor >= 0)
                {
                    pCurDst->r = 4 * m_Hdr.m_pPal[nColor];
                    pCurDst->g = 4 * m_Hdr.m_pPal[nColor + 1];
                    pCurDst->b = 4 * m_Hdr.m_pPal[nColor + 2];
                    pCurDst->a = 255;
                }

                ++pCurSrc;
                ++pCurDst;
            }
        }
        
        nCurX += m_Hdr.m_nCharSpace + Char.m_uShift;
    }
}
//---------------------------------------------------------------------------
bool BmfFntImpl_t::LoadData(const char *pszFile, Data_t &Data) const
{
    FS_VERIFY_RETURN_VAL(pszFile, false);

    // open the BMF file
    FILE *pIn = ::fopen(pszFile, "rb");
    if(!pIn)
        return false; // couldn't open the file
        
    // get the size of the file
    FS_VERIFY(::fseek(pIn, 0, SEEK_END) == 0);
    long nFileSize = ::ftell(pIn);
    FS_VERIFY(::fseek(pIn, 0, SEEK_SET) == 0);

    // read the whole content    
    Data.resize(nFileSize);
    FS_VERIFY(::fread(&Data[0], nFileSize, 1, pIn) == 1);
    
    ::fclose(pIn);
    return true;
}
//---------------------------------------------------------------------------
void BmfFntImpl_t::BindChars()
{
    m_Chars.resize(256);
            
    const Hdr_t::uint8 *pCharStart = m_Hdr.m_pChars;
    for(int i = 0; i < m_Hdr.m_uNumChars; ++i)
    {
        Char_t Char(pCharStart);
        
        FS_ASSERT(Char.m_uID >= 0 && Char.m_uID < 256);
        m_Chars[Char.m_uID] = Char;
        
        pCharStart = Char.m_pData + Char.m_uW * Char.m_uH;
    }
}
//---------------------------------------------------------------------------
BmfFntImpl_t::Hdr_t::Hdr_t(const uint8 *pData):
m_bValid(true),
m_uMagic(*reinterpret_cast<const uint32 *>(pData)),
m_uVer(*(pData + 4)),
m_uLnH(*(pData + 5)),
m_nOverLn(*reinterpret_cast<const int8 *>(pData + 6)),
m_nUnderLn(*reinterpret_cast<const int8 *>(pData + 7)),
m_nCharSpace(*reinterpret_cast<const int8 *>(pData + 8)),
m_nInnerSize(*reinterpret_cast<const int8 *>(pData + 9)),
m_uNumClrs(*(pData + 10)),
m_uMaxClr(*(pData + 11)),
m_uRsrvd(*(pData + 12)),
m_uNumRgb(*(pData + 16)),
m_pPal(pData + 17),
m_uInfoLen(*(pData + 17 + 3 * m_uNumRgb)),
m_pInfo(pData + 18 + 3 * m_uNumRgb),
m_uNumChars(*reinterpret_cast<const uint8 *>(pData + 18 +
 3 * m_uNumRgb + m_uInfoLen)),
m_pChars(pData + 20 + 3 * m_uNumRgb + m_uInfoLen)
{
    FS_STATIC_ASSERT(sizeof(uint32) == 4);
}
//---------------------------------------------------------------------------
BmfFntImpl_t::Char_t::Char_t(const uint8 *pData):
m_bValid(true),
m_uID(*pData),
m_uW(*(pData + 1)),
m_uH(*(pData + 2)),
m_nRelX(*reinterpret_cast<const int8 *>(pData + 3)),
m_nRelY(*reinterpret_cast<const int8 *>(pData + 4)),
m_uShift(*(pData + 5)),
m_pData(pData + 6)
{
}
//---------------------------------------------------------------------------
} // anonymous namespace
//---------------------------------------------------------------------------
//--- CLASS BmfFnt_t >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
//---------------------------------------------------------------------------
BmfFnt_t::BmfFnt_t(const char *pszFile):
m_pImpl(new BmfFntImpl_t(pszFile))
{
    FS_ASSERT_MSG(m_pImpl, "Invalid implementation pointer!");
}
//---------------------------------------------------------------------------
BmfFnt_t::~BmfFnt_t()
{
    FS_ASSERT_MSG(m_pImpl, "Invalid implementation pointer!");
    delete static_cast<BmfFntImpl_t *>(m_pImpl);
}
//---------------------------------------------------------------------------
bool BmfFnt_t::IsValid() const
{
    FS_ASSERT_MSG(m_pImpl, "Invalid implementation pointer!");
    return static_cast<BmfFntImpl_t *>(m_pImpl)->IsValid();
}
//---------------------------------------------------------------------------
void BmfFnt_t::Print(int nX, int nY, const char *pszText, PxFmt8888_t
 *pDst, int nDstW, int nDstH, int nClipX1, int nClipY1, int nClipX2,
 int nClipY2) const
{
    FS_ASSERT_MSG(m_pImpl, "Invalid implementation pointer!");
    
    static_cast<BmfFntImpl_t *>(m_pImpl)->Print(nX, nY, pszText, pDst,
     nDstW, nDstH, nClipX1, nClipY1, nClipX2, nClipY2);
}
//---------------------------------------------------------------------------