/*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*/
/*                          _______      ______    ______          __      */
/*  ~ ~ ~ ~ ~ ~ ~ ~ ~ ~    / ____(_)___ / ___ /)  / ____/___  ____/ /__    */
/*     [fsBmfFnt]         / /_  / / __ \\__ \|/  / /   / __ \/ __  / _ \   */
/*       rev. 3          / __/ / / /_/ /__/ /   / /___/ /_/ / /_/ /  __/   */
/*    14th Feb 2007     /_/   /_/ ,___/____/    \____/\____/\__,_/\___/    */
/*     [ ] stable              /_/ (c) 2006-7 Filip STOKLAS (FipS)         */
/*  ~ ~ ~ ~ ~ ~ ~ ~ ~ ~       http://HOLE.4FipS.com/fips_code.php          */
/*                                                                         */
/* This code is free for personal and commercial use. You may redistribute */
/* it by any means. If you use the code for your own projects please give  */
/* me credit. Please send a bug report. Don't alter or remove this header! */
/*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*/
#ifndef FS_RAS_FSBMFFNT_H
#define FS_RAS_FSBMFFNT_H
//---------------------------------------------------------------------------
#if defined(FS_INCLUDE_USERDEFS) // inject user definition file on request
#   include <fs/fsUserDefs.h>
#endif 
//---------------------------------------------------------------------------
namespace fs { namespace ras {
//---------------------------------------------------------------------------
//! BMF font loader / renderer.
/*! After a BMF font is loaded, a text string using the font can be rendered
    into the destination bitmap surface. Many free BMF fonts can be
    found <a href="http://bmf.wz.cz/">here</a>. There is also BMF format
    <a href="http://bmf.wz.cz/bmf-format.htm"> specification</a>.
    
    \image html fs_ras_BmfFnt_t_01.gif
    
    \todo Clipping, RGB-888 support, Consider splitting.            
*/
class BmfFnt_t
{
 public:

    //! RGB-888 destination bitmap surface pixel format.
    struct PxFmt888_t { unsigned char r, g, b; };
    //! RGBA-8888 destination bitmap surface pixel format.
    struct PxFmt8888_t { unsigned char r, g, b, a; };
    
    //! Loads a BMF font using the given filename.
    explicit BmfFnt_t(const char *pszFile);
    ~BmfFnt_t();
    //! Checks the internal state / correct construction.
    bool IsValid() const;
    
    //! Render a text string into the destination bitmap surface.
    void Print(int nX, int nY, const char *pszText, PxFmt8888_t *pDst,
     int nDstW, int nDstH, int nClipX1 = -1, int nClipY1 = -1,
     int nClipX2 = -1, int nClipY2 = -1) const;

 private:
 
    BmfFnt_t(); //!< No def ctor.
    BmfFnt_t(const BmfFnt_t &); //!< No copy.
    BmfFnt_t & operator = (const BmfFnt_t &); //!< No assignment.
    
    void *m_pImpl; //!< The PIPML idiom is used.
};
//---------------------------------------------------------------------------
}} // namespace fs::ras
//---------------------------------------------------------------------------
#endif // FS_RAS_FSBMFFNT_H