/*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*/
/*                          _______      ______    ______          __      */
/*  ~ ~ ~ ~ ~ ~ ~ ~ ~ ~    / ____(_)___ / ___ /)  / ____/___  ____/ /__    */
/*     [fsRle8Enc]        / /_  / / __ \\__ \|/  / /   / __ \/ __  / _ \   */
/*       rev. 6          / __/ / / /_/ /__/ /   / /___/ /_/ / /_/ /  __/   */
/*    6th Feb 2007      /_/   /_/ ,___/____/    \____/\____/\__,_/\___/    */
/*     [x] stable              /_/ (c) 2006-7 Filip STOKLAS (FipS)         */
/*  ~ ~ ~ ~ ~ ~ ~ ~ ~ ~       http://HOLE.4FipS.com/fips_code.php          */
/*                                                                         */
/* This code is free for personal and commercial use. You may redistribute */
/* it by any means. If you use the code for your own projects please give  */
/* me credit. Please send a bug report. Don't alter or remove this header! */
/*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*/
#ifndef FS_RAS_FSRLE8ENC_H
#define FS_RAS_FSRLE8ENC_H
//---------------------------------------------------------------------------
#if defined(FS_INCLUDE_USERDEFS) // inject user definition file on request
#   include <fs/fsUserDefs.h>
#endif 
//---------------------------------------------------------------------------
namespace fs { namespace ras {
//---------------------------------------------------------------------------
//! RLE8 encoder.
/*! Rle8Enc_t represents a <b>RLE8 encoder</b>. Source <b>RGBA-8888</b>
    image data is encoded into <b>sequences of opaque pixels</b> referring
    to a <b>RGBA-8888 color palette</b> through an 8bit index. Transparent
    segments are left out. Each chunk in the RLE8 data stream starts with a
    <b>[COUNT] byte</b> holding number of pixels (max. <b>127</b>) in the
    subsequent sequence. <b>[COUNT]</b> has also encoded a <b>transparency
    flag</b> implying whether or not an opaque pixel sequence follows. See
    the following example for a basic idea: <i>0,0,0,1,2,3,0,0,5,0,0,0 is
    encoded into: 3/0,3/1,1,2,3,2/0,1/1,5,3/0</i>. Note that the <b>full
    width is always encoded</b> even if there are only transparent pixels
    on a scan line.
*/
class Rle8Enc_t
{
 public:

    typedef unsigned int uint32;
    struct PxFmt8888_t //! Input & palette pixel format.
    {
        unsigned char r, g, b, a;
        bool operator == (const PxFmt8888_t &RHS)
        { return r == RHS.r && g == RHS.g && b == RHS.b && a == RHS.a; }
    };

    //! Initialize from the given RGBA-8888 source image data.
    Rle8Enc_t(const PxFmt8888_t *pSrc, int nW, int nH);
    //! Destroys internal data.
    ~Rle8Enc_t();
    //! Checks the internal state / correct construction.
    bool IsValid() const;

    //! Returns the size of the image.
    bool GetSize(int &nW, int &nH) const;
    //! Returns RLE8 data stream.
    bool GetRle(const unsigned char *&pRle, int &nNum) const;
    //! Returns palette items.
    bool GetPal(const PxFmt8888_t *&pPal, int &nNum) const;
    //! Returns line offsets.
    bool GetOff(const uint32 *&pOff, int &nNum) const;

 private:
 
    Rle8Enc_t(); //!< No def ctor.
    Rle8Enc_t(const Rle8Enc_t &); //!< No copy.
    Rle8Enc_t & operator = (const Rle8Enc_t &); //!< No assignment.
 
    void *m_pImpl; //!< The PIPML idiom is used.
};
//---------------------------------------------------------------------------
}} // namespace fs::ras
//---------------------------------------------------------------------------
#endif // FS_RAS_FSRLE8ENC_H