/*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*/
/*                          _______      ______    ______          __      */
/*  ~ ~ ~ ~ ~ ~ ~ ~ ~ ~    / ____(_)___ / ___ /)  / ____/___  ____/ /__    */
/*     [fsTgaLdr]         / /_  / / __ \\__ \|/  / /   / __ \/ __  / _ \   */
/*       rev. 4          / __/ / / /_/ /__/ /   / /___/ /_/ / /_/ /  __/   */
/*    5th Feb 2007      /_/   /_/ ,___/____/    \____/\____/\__,_/\___/    */
/*     [x] stable              /_/ (c) 2006-7 Filip STOKLAS (FipS)         */
/*  ~ ~ ~ ~ ~ ~ ~ ~ ~ ~       http://HOLE.4FipS.com/fips_code.php          */
/*                                                                         */
/* This code is free for personal and commercial use. You may redistribute */
/* it by any means. If you use the code for your own projects please give  */
/* me credit. Please send a bug report. Don't alter or remove this header! */
/*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*/
#ifndef FS_RAS_FSTGALDR_H
#define FS_RAS_FSTGALDR_H
//---------------------------------------------------------------------------
#if defined(FS_INCLUDE_USERDEFS) // inject user definition file on request
#   include <fs/fsUserDefs.h>
#endif
//---------------------------------------------------------------------------
namespace fs { namespace ras {
//---------------------------------------------------------------------------
//! TGA image loader / data holder.
/*! The class loads a <b>TGA</b> image file and then holds its image data.
    Currently, two formats <b>RGB-888</b> and <b>RGBA-8888</b> are
    supported. After loading, the data is <b>rearranged</b> if necessary
    to ensure <b>R-G-B-(A)</b> channel order.
*/
class TgaLdr_t
{
 public:

    //! TgaLdr_t : RGB-888 pixel format.
    struct PxFmt888_t { unsigned char r, g, b; };
    //! TgaLdr_t : RGBA-8888 pixel format. 
    struct PxFmt8888_t { unsigned char r, g, b, a; };

    //! Loads a TGA file using the given filename.
    explicit TgaLdr_t(const char *pszFile);
    ~TgaLdr_t();
    //! Checks the internal state / correct construction.
    bool IsValid() const;

    //! Returns the width of the image.
    int GetW() const;
    //! Returns the height of the image.
    int GetH() const;
    //! Returns a pointer to RGB-888 data or 0 if the format doesn't match.
    const PxFmt888_t * GetData888() const;
    //! Returns a pointer to RGBA-8888 data or 0 if the format doesn't match.
    const PxFmt8888_t * GetData8888() const;

 private:

    TgaLdr_t(); //!< No def ctor.
    TgaLdr_t(const TgaLdr_t &); //!< No copy.
    TgaLdr_t & operator = (const TgaLdr_t &); //!< No assignment.

    void *m_pImpl; //!< The PIPML idiom is used.
};
//---------------------------------------------------------------------------
}} // namespace fs::ras
//---------------------------------------------------------------------------
#endif // FS_RAS_FSTGALDR_H