/*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*/
/*                          _______      ______    ______          __      */
/*  ~ ~ ~ ~ ~ ~ ~ ~ ~ ~    / ____(_)___ / ___ /)  / ____/___  ____/ /__    */
/*       [fsDll]          / /_  / / __ \\__ \|/  / /   / __ \/ __  / _ \   */
/*       rev. 8          / __/ / / /_/ /__/ /   / /___/ /_/ / /_/ /  __/   */
/*    10th Jan 2007     /_/   /_/ ,___/____/    \____/\____/\__,_/\___/    */
/*     [x] stable              /_/ (c) 2004-7 Filip STOKLAS (FipS)         */
/*  ~ ~ ~ ~ ~ ~ ~ ~ ~ ~       http://HOLE.4FipS.com/fips_code.php          */
/*                                                                         */
/* This code is free for personal and commercial use. You may redistribute */
/* it by any means. If you use the code for your own projects please give  */
/* me credit. Please send a bug report. Don't alter or remove this header! */
/*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*/
#include <fs/sys/fsDll.h>

#if defined(UNDER_CE)
#   include <fs/ut/fsStrUt.h> // wide-character string utilities
#endif

// >>> COMMON SUPPORT BLOCK (rev. 6)
#if defined(FS_HAS_FSASSERT)
#   include <fs/sys/fsAssert.h>
#else
#   include <cassert>
#   if !defined(FS_ASSERT)
#       define FS_ASSERT(exp) assert(exp)
#   endif
#   if !defined(FS_ASSERT_MSG)
#       define FS_ASSERT_MSG(exp, msg) assert(exp && msg)
#   endif
#   if !defined(FS_VERIFY)
#       define FS_VERIFY(exp) \
        { bool bExp = !!(exp); assert(bExp && #exp); bExp; }
#   endif
#   if !defined(FS_VERIFY_MSG)
#       define FS_VERIFY_MSG(exp, msg) \
        { bool bExp = !!(exp); assert(bExp && #exp && msg); bExp; }
#   endif
#   if !defined(FS_STATIC_ASSERT)
#       define FS_STATIC_ASSERT(exp) { char error[(exp) ? 1 : 0]; error; }
#   endif
#endif
#if !defined(FS_VERIFY_RETURN)
#   define FS_VERIFY_RETURN(exp) \
    if(!(exp)) { FS_ASSERT_MSG(0, #exp "<return>"); return; }
#endif
#if !defined(FS_VERIFY_RETURN_VAL)
#   define FS_VERIFY_RETURN_VAL(exp, ret) \
    if(!(exp)) { FS_ASSERT_MSG(0, #exp "<returns:>" #ret); return (ret); }
#endif
// <<< COMMON SUPPORT BLOCK (rev. 6)

#include <string>

#if defined(_WIN32)
#   include <windows.h>
#elif defined(FS_PLATFORM_LINUX)
#   include <dlfcn.h>
#endif

// provide detailed memory tracking information
#if defined(FS_HAS_FSMEMMGR)
#   include <fs/sys/fsMemMgr.h>
#endif
//---------------------------------------------------------------------------
using namespace fs::sys;
//---------------------------------------------------------------------------
//--- ANONYMOUS NAMESPACE >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
//---------------------------------------------------------------------------
namespace {
//---------------------------------------------------------------------------
//--- CLASS DllImpl_t >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
//---------------------------------------------------------------------------
class DllImpl_t
{
public:

    explicit DllImpl_t(const char *pszFile);
    ~DllImpl_t();
    bool IsValid() const;
        
    const void * GetSymb(const char *pszName) const;
    const char * GetFile() const;

private:

    DllImpl_t();
    DllImpl_t(const DllImpl_t &);
    DllImpl_t & operator = (const DllImpl_t &);
    
    void * Load(const char *pszFile) const;

    void *m_pHandle;
    std::string m_strFile;
};
//---------------------------------------------------------------------------
DllImpl_t::DllImpl_t(const char *pszFile):
m_pHandle(0)
{
    void *pHandle = Load(pszFile);
    FS_VERIFY_RETURN(pHandle);
    
    //--- TRANSFER THE STATE --- 
    
    m_pHandle = pHandle;
    m_strFile = pszFile;
}
//---------------------------------------------------------------------------
DllImpl_t::~DllImpl_t()
{
#if defined(_WIN32)

    if(m_pHandle)
        ::FreeLibrary(static_cast<HINSTANCE>(m_pHandle));

#elif defined(FS_PLATFORM_LINUX)

    if(m_pHandle)
        ::dlclose(m_pHandle);

#else
#   error Unknown platform!
#endif
}
//---------------------------------------------------------------------------
bool DllImpl_t::IsValid() const
{
    return m_pHandle ? true : false;
}
//---------------------------------------------------------------------------
const void * DllImpl_t::GetSymb(const char *pszName) const
{
    FS_VERIFY_RETURN_VAL(IsValid(), 0);
    FS_VERIFY_RETURN_VAL(pszName, 0);
    FS_ASSERT(m_pHandle);
    
#if defined(UNDER_CE)
    
    return ::GetProcAddress(static_cast<HINSTANCE>(m_pHandle),
     fs::ut::str::StrHldr_t(pszName).GetStr());

#elif defined(_WIN32)

    return ::GetProcAddress(static_cast<HINSTANCE>(m_pHandle), pszName);
    
#elif defined(FS_PLATFORM_LINUX)

    return ::dlsym(m_pHandle, pszName);
    
#else
#   error Unknown platform!
#endif
}
//---------------------------------------------------------------------------
const char * DllImpl_t::GetFile() const
{
    FS_VERIFY_RETURN_VAL(IsValid(), 0);
    
    return m_strFile.c_str();
}
//---------------------------------------------------------------------------
void * DllImpl_t::Load(const char *pszFile) const
{
    FS_VERIFY_RETURN_VAL(pszFile, 0);
    
    void *pHandle = 0;
    
#if defined(UNDER_CE)

    pHandle = ::LoadLibrary(fs::ut::str::WStrHldr_t(pszFile).GetWStr());

#elif defined(_WIN32)

    pHandle = ::LoadLibraryA(pszFile);

#elif defined(FS_PLATFORM_LINUX)

    pHandle = ::dlopen(pszFile, RTLD_LAZY);

#else
#   error Unknown platform!
#endif

    return pHandle;
}
//---------------------------------------------------------------------------
} // anonymous namespace
//---------------------------------------------------------------------------
//--- CLASS Dll_t >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
//---------------------------------------------------------------------------
Dll_t::Dll_t(const char *pszFile):
m_pImpl(new DllImpl_t(pszFile))
{
    FS_ASSERT_MSG(m_pImpl, "Invalid implementation pointer!");
}
//---------------------------------------------------------------------------
Dll_t::~Dll_t()
{
    FS_ASSERT_MSG(m_pImpl, "Invalid implementation pointer!");
    delete static_cast<DllImpl_t *>(m_pImpl);
}
//---------------------------------------------------------------------------
bool Dll_t::IsValid() const
{
    FS_ASSERT_MSG(m_pImpl, "Invalid implementation pointer!");
    return static_cast<DllImpl_t *>(m_pImpl)->IsValid();
}
//---------------------------------------------------------------------------
const void * Dll_t::GetSymb(const char *pszName) const
{
    FS_ASSERT_MSG(m_pImpl, "Invalid implementation pointer!");
    return static_cast<DllImpl_t *>(m_pImpl)->GetSymb((pszName));
}
//---------------------------------------------------------------------------
const char * Dll_t::GetFile() const
{
    FS_ASSERT_MSG(m_pImpl, "Invalid implementation pointer!");
    return static_cast<DllImpl_t *>(m_pImpl)->GetFile();
}
//---------------------------------------------------------------------------