/*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*/
/*                          _______      ______    ______          __      */
/*  ~ ~ ~ ~ ~ ~ ~ ~ ~ ~    / ____(_)___ / ___ /)  / ____/___  ____/ /__    */
/*       [fsDll]          / /_  / / __ \\__ \|/  / /   / __ \/ __  / _ \   */
/*       rev. 8          / __/ / / /_/ /__/ /   / /___/ /_/ / /_/ /  __/   */
/*    10th Jan 2007     /_/   /_/ ,___/____/    \____/\____/\__,_/\___/    */
/*     [x] stable              /_/ (c) 2004-7 Filip STOKLAS (FipS)         */
/*  ~ ~ ~ ~ ~ ~ ~ ~ ~ ~       http://HOLE.4FipS.com/fips_code.php          */
/*                                                                         */
/* This code is free for personal and commercial use. You may redistribute */
/* it by any means. If you use the code for your own projects please give  */
/* me credit. Please send a bug report. Don't alter or remove this header! */
/*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*/
#ifndef FS_SYS_FSDLL_H
#define FS_SYS_FSDLL_H
//---------------------------------------------------------------------------
#if defined(FS_INCLUDE_USERDEFS) // inject user definition file on request
#   include <fs/fsUserDefs.h>
#endif
//---------------------------------------------------------------------------
namespace fs { namespace sys {
//---------------------------------------------------------------------------
//! System DLL wrapper.
/*! Dll_t wraps a system <b>dynamic-link library</b>. It helps manage
    the <b>lifetime</b> of the library (it follows <b>RAII</b> pattern),
    the library is <b>automatically unloaded</b> when Dll_t is destroyed.
*/
class Dll_t
{
 public:
 
    //! Loads the system DLL using the given filename.
    explicit Dll_t(const char *pszFile);
    //! Unloads the system DLL.
    ~Dll_t();
    //! Checks the internal state / correct construction.
    bool IsValid() const;
    
    //! Returns the address of the specified exported DLL symbol.
    const void * GetSymb(const char *pszName) const;
    //! Returns the filename of the loaded DLL.
    const char * GetFile() const;

 private:
 
    Dll_t(); //!< No def ctor.
    Dll_t(const Dll_t &); //!< No copy.
    Dll_t & operator = (const Dll_t &); //!< No assignment.
 
    void *m_pImpl; //!< The PIPML idiom is used.
};
//---------------------------------------------------------------------------
}} // namespace fs::sys
//---------------------------------------------------------------------------
#endif // FS_SYS_FSDLL_H