/*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*/
/*                          _______      ______    ______          __      */
/*  ~ ~ ~ ~ ~ ~ ~ ~ ~ ~    / ____(_)___ / ___ /)  / ____/___  ____/ /__    */
/*     [fsFileDir]        / /_  / / __ \\__ \|/  / /   / __ \/ __  / _ \   */
/*       rev. 8          / __/ / / /_/ /__/ /   / /___/ /_/ / /_/ /  __/   */
/*    15th Jan 2007     /_/   /_/ ,___/____/    \____/\____/\__,_/\___/    */
/*     [x] stable              /_/ (c) 2004-07 Filip STOKLAS (FipS)        */
/*  ~ ~ ~ ~ ~ ~ ~ ~ ~ ~       http://HOLE.4FipS.com/fips_code.php          */
/*                                                                         */
/* This code is free for personal and commercial use. You may redistribute */
/* it by any means. If you use the code for your own projects please give  */
/* me credit. Please send a bug report. Don't alter or remove this header! */
/*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*/
#include <fs/sys/fsFileDir.h>

#if defined(UNDER_CE)
#   include <fs/ut/fsStrUt.h> // wide-character string helpers
#endif

// >>> COMMON SUPPORT BLOCK (rev. 6)
#if defined(FS_HAS_FSASSERT)
#   include <fs/sys/fsAssert.h>
#else
#   include <cassert>
#   if !defined(FS_ASSERT)
#       define FS_ASSERT(exp) assert(exp)
#   endif
#   if !defined(FS_ASSERT_MSG)
#       define FS_ASSERT_MSG(exp, msg) assert(exp && msg)
#   endif
#   if !defined(FS_VERIFY)
#       define FS_VERIFY(exp) \
        { bool bExp = !!(exp); assert(bExp && #exp); bExp; }
#   endif
#   if !defined(FS_VERIFY_MSG)
#       define FS_VERIFY_MSG(exp, msg) \
        { bool bExp = !!(exp); assert(bExp && #exp && msg); bExp; }
#   endif
#   if !defined(FS_STATIC_ASSERT)
#       define FS_STATIC_ASSERT(exp) { char error[(exp) ? 1 : 0]; error; }
#   endif
#endif
#if !defined(FS_VERIFY_RETURN)
#   define FS_VERIFY_RETURN(exp) \
    if(!(exp)) { FS_ASSERT_MSG(0, #exp "<return>"); return; }
#endif
#if !defined(FS_VERIFY_RETURN_VAL)
#   define FS_VERIFY_RETURN_VAL(exp, ret) \
    if(!(exp)) { FS_ASSERT_MSG(0, #exp "<returns:>" #ret); return (ret); }
#endif
// <<< COMMON SUPPORT BLOCK (rev. 6)

#include <vector>
#include <string>

#if defined(_WIN32)
#   include <windows.h>
#endif

// provide detailed memory tracking information
#if defined(FS_HAS_FSMEMMGR)
#   include <fs/sys/fsMemMgr.h>
#endif
//---------------------------------------------------------------------------
#if (_MSC_VER >= 1400) // MSVC8
#   pragma warning(disable : 4996) // function was marked as deprecated
#endif
//---------------------------------------------------------------------------
using namespace fs::sys::fdr;
//---------------------------------------------------------------------------
//--- FUNCTIONS >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
//---------------------------------------------------------------------------
const char * fs::sys::fdr::GetPathSpr()
{
#if defined(_WIN32)

    static char szSpr[] = "\\";
    return szSpr;
    
#else
#   error Unknown platform!
#endif
}
//---------------------------------------------------------------------------
const char * fs::sys::fdr::GetNewLn()
{
    static char szNewLn[] = "\n";
    return szNewLn;
}
//---------------------------------------------------------------------------
//! \todo Make the whole body static.
const char * fs::sys::fdr::GetAppDir()
{
#if defined(UNDER_CE)

    static wchar_t wszTemp[MAX_PATH + 1];
    ::GetModuleFileName(0, wszTemp, MAX_PATH);
    
    static char szTemp[MAX_PATH + 1];
    fs::ut::str::WStr2Str(wszTemp, szTemp);
    
    char *pszSprPos = ::strrchr(szTemp, GetPathSpr()[0]);
    FS_ASSERT(pszSprPos);
    *(pszSprPos + 1) = '\0';
    
    return szTemp;

#elif defined(_WIN32)
    
    static char szTemp[MAX_PATH + 1];
    ::GetCurrentDirectoryA(MAX_PATH, szTemp);
    
    ::strcat(szTemp, GetPathSpr());
    
    return szTemp;

#else
#   error Unknown platform!
#endif
}
//---------------------------------------------------------------------------
bool fs::sys::fdr::IsDirExist(const char *pszFullDir)
{
    FS_VERIFY_RETURN_VAL(pszFullDir, false);
    
#if defined(UNDER_CE)

    using namespace fs::ut::str;
    DWORD Attr = ::GetFileAttributes(WStrHldr_t(pszFullDir).GetWStr());
    
    return Attr != 0xFFFFFFFF && (Attr | FILE_ATTRIBUTE_DIRECTORY);

#elif defined(_WIN32)

    DWORD Attr = ::GetFileAttributesA(pszFullDir);
    
    return Attr != INVALID_FILE_ATTRIBUTES && (Attr |
     FILE_ATTRIBUTE_DIRECTORY);

#else
#   error Unknown platform!
#endif
}
//---------------------------------------------------------------------------
/*! \param pszFullDir A valid system dependent full directory path is
     expected, no internal checking is performed.
    \return <b>true</b> indicates success. Returns <b>false</b> if it was
     not possible to create the requested directory structure completely.
*/
bool fs::sys::fdr::MakeDir(const char *pszFullDir)
{
    FS_VERIFY_RETURN_VAL(pszFullDir, false);
    
#if defined(UNDER_CE)

    FS_ASSERT_MSG(0, "Not implemented!");
    return false;

#elif defined(_WIN32)

    const char chSpr = GetPathSpr()[0];
    const char *pSpr = ::strchr(pszFullDir, chSpr);
    FS_ASSERT(pSpr);
    
    do
    {
        std::string strSub(pszFullDir, 0, pSpr - pszFullDir + 1);
                
        if(!IsDirExist(strSub.c_str()))
            if(!::CreateDirectoryA(strSub.c_str(), 0))
                return false;
            
        pSpr = ::strchr(pSpr + 1, chSpr);
    }
    while(pSpr);
    
    return true;

#else
#   error Unknown platform!
#endif
}
//---------------------------------------------------------------------------
/*! If there is no path separator at the end of the first part, it is
    added, then the second part (which is optional) is appended. Path
    separators are replaced according to current environment configuration
    and multiple separators are removed.
    \note Be sure the size of pszOut is big enough for holding the
     resulting string, which is always shorter then combined original
     parts.
*/
void fs::sys::fdr::NicePath(char *pszOut, const char *pszDir, const char
 *pszFile)
{
    FS_VERIFY_RETURN(pszOut && pszDir);

    // form the directory part
    
    std::string str = std::string(pszDir) + std::string(GetPathSpr());
    
    if(pszFile) // the file part is optional
        str.append(pszFile);
        
    // remove multiple path separators and swap '/' <=> '\'
    
    const char chSpr = GetPathSpr()[0];
    const char chBadSpr = chSpr == '/' ? '\\' : '/';
    
    std::string::iterator it = str.begin();
    char chOld = '\0';
    
    while(it != str.end())
    {
        char &ch = *it;
    
        if(ch == chBadSpr) // '/' <=> '\'
            ch = chSpr;
            
        // eliminate multiple path separators
        if(ch == chSpr && chOld == chSpr)
        {
            str.erase(it);
            continue;
        }

        chOld = ch;
        it++;
    }
    
    ::strcpy(pszOut, str.c_str());
}
//---------------------------------------------------------------------------
//--- ANONYMOUS NAMESPACE >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
//---------------------------------------------------------------------------
namespace {
//---------------------------------------------------------------------------
//--- CLASS DirLstImpl_t >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
//---------------------------------------------------------------------------
class DirLstImpl_t
{
public:

    explicit DirLstImpl_t(const char *pszFullDir, bool bRecur = false,
     bool bInclDirs = false, bool bInclFiles = true);
    bool IsValid() const;

    const char * GetRoot() const;
    int GetNum() const;
    const char * GetAt(int nIndex) const;

private:

    DirLstImpl_t();
    DirLstImpl_t(const DirLstImpl_t &);
    DirLstImpl_t & operator = (const DirLstImpl_t &);
    
    //---
    
    std::string m_strRoot;
    std::vector<std::string> m_Items;
};
//---------------------------------------------------------------------------
DirLstImpl_t::DirLstImpl_t(const char *pszFullDir, bool bRecur,
 bool bInclDirs, bool bInclFiles):
m_strRoot(pszFullDir ? pszFullDir : "")
{
    FS_VERIFY_RETURN(pszFullDir);

#if defined(_WIN32)

    const char * const pSpr = fs::sys::fdr::GetPathSpr();

    //--- LIST DIRECTORIES ---
    
    if(bRecur)
    {
        std::string strRoot = pszFullDir;
                        
        int nPos = -1;
        while(nPos < static_cast<int>(m_Items.size()))
        {
            if(nPos >= 0)
                strRoot = std::string(pszFullDir) + m_Items[nPos];
                
            strRoot.append("*.*");
            
            WIN32_FIND_DATAA FindData;
            HANDLE hFile = ::FindFirstFileA(strRoot.c_str(), &FindData);
            bool bDone = hFile == INVALID_HANDLE_VALUE;
        
            while(!bDone)
            {
                if(FindData.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY &&
                 ::strcmp(FindData.cFileName, ".") != 0 &&
                 ::strcmp(FindData.cFileName, "..") != 0
                )
                {
                    m_Items.push_back(
                     std::string(nPos >= 0 ? m_Items[nPos] : "").
                     append(FindData.cFileName).append(pSpr)
                    );
                }
                
                bDone = ::FindNextFileA(hFile, &FindData) ? false : true;
            }
            
            ::FindClose(hFile);
            ++nPos;
        }
    }
    
    const int nDirCount = static_cast<int>(m_Items.size());
        
    //--- LIST FILES ---
    
    if(bInclFiles)
    {
        std::string strRoot = pszFullDir;
        
        int nPos = -1;
                
        while(nPos < nDirCount)
        {
            if(nPos >= 0)
                strRoot = std::string(pszFullDir) + m_Items[nPos];
                
            strRoot.append("*.*");
            
            WIN32_FIND_DATAA FindData;
            HANDLE hFile = ::FindFirstFileA(strRoot.c_str(), &FindData);
            bool bDone = hFile == INVALID_HANDLE_VALUE;
        
            while(!bDone)
            {
                if(!(FindData.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY))
                {
                    m_Items.push_back(
                     std::string(nPos >= 0 ? m_Items[nPos] : "").
                     append(FindData.cFileName)
                    );
                }
                
                bDone = ::FindNextFileA(hFile, &FindData) ? false : true;
            }
            
            ::FindClose(hFile);
            ++nPos;
        }
    }
    
    //--- REMOVE UNNECESSARY DIRECTORIES ---    
    
    if(!bInclDirs)
        m_Items.erase(m_Items.begin(), m_Items.begin() + nDirCount);
        
    //---
            
#else
#   error Unknown platform!
#endif
}
//---------------------------------------------------------------------------
bool DirLstImpl_t::IsValid() const
{
    return !m_strRoot.empty();
}
//---------------------------------------------------------------------------
const char * DirLstImpl_t::GetRoot() const
{
    FS_VERIFY_RETURN_VAL(IsValid(), 0);

    return m_strRoot.c_str();
}
//---------------------------------------------------------------------------
int DirLstImpl_t::GetNum() const
{
    FS_VERIFY_RETURN_VAL(IsValid(), 0);

    return static_cast<int>(m_Items.size());
}
//---------------------------------------------------------------------------
const char * DirLstImpl_t::GetAt(int nIndex) const
{
    FS_VERIFY_RETURN_VAL(IsValid(), 0);
    FS_VERIFY_RETURN_VAL(nIndex >= 0 && nIndex < GetNum(), 0);
    
    return m_Items[nIndex].c_str();
}
//---------------------------------------------------------------------------
} // anonymous namespace
//---------------------------------------------------------------------------
//--- CLASS DirLst_t >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
//---------------------------------------------------------------------------
/*! \param pszFullDir A valid system dependent full directory path is
     expected, no internal checking is performed.
    \param bRecur Specifies whether subdirectories should be also
     traversed recursively.
    \param bInclFiles Specifies whether files should be also included
     in the list.
    \param bInclDirs Specifies whether directories should be also
     included in the list.
*/
DirLst_t::DirLst_t(const char *pszFullDir, bool bRecur, bool bInclDirs,
 bool bInclFiles):
m_pImpl(new DirLstImpl_t(pszFullDir, bRecur, bInclDirs, bInclFiles))
{
    FS_ASSERT_MSG(m_pImpl, "Invalid implementation pointer!");
}
//---------------------------------------------------------------------------
DirLst_t::~DirLst_t()
{
    FS_ASSERT_MSG(m_pImpl, "Invalid implementation pointer!");
    delete static_cast<DirLstImpl_t *>(m_pImpl);
}
//---------------------------------------------------------------------------
bool DirLst_t::IsValid() const
{
    FS_ASSERT_MSG(m_pImpl, "Invalid implementation pointer!");
    return static_cast<DirLstImpl_t *>(m_pImpl)->IsValid();
}
//---------------------------------------------------------------------------
const char * DirLst_t::GetRoot() const
{
    FS_ASSERT_MSG(m_pImpl, "Invalid implementation pointer!");
    return static_cast<DirLstImpl_t *>(m_pImpl)->GetRoot();
}
//---------------------------------------------------------------------------
int DirLst_t::GetNum() const
{
    FS_ASSERT_MSG(m_pImpl, "Invalid implementation pointer!");
    return static_cast<DirLstImpl_t *>(m_pImpl)->GetNum();
}
//---------------------------------------------------------------------------
const char * DirLst_t::GetAt(int nIndex) const
{
    FS_ASSERT_MSG(m_pImpl, "Invalid implementation pointer!");
    return static_cast<DirLstImpl_t *>(m_pImpl)->GetAt(nIndex);
}
//---------------------------------------------------------------------------