/*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*/
/*                          _______      ______    ______          __      */
/*  ~ ~ ~ ~ ~ ~ ~ ~ ~ ~    / ____(_)___ / ___ /)  / ____/___  ____/ /__    */
/*     [fsFileDir]        / /_  / / __ \\__ \|/  / /   / __ \/ __  / _ \   */
/*       rev. 8          / __/ / / /_/ /__/ /   / /___/ /_/ / /_/ /  __/   */
/*    15th Jan 2007     /_/   /_/ ,___/____/    \____/\____/\__,_/\___/    */
/*     [x] stable              /_/ (c) 2004-07 Filip STOKLAS (FipS)        */
/*  ~ ~ ~ ~ ~ ~ ~ ~ ~ ~       http://HOLE.4FipS.com/fips_code.php          */
/*                                                                         */
/* This code is free for personal and commercial use. You may redistribute */
/* it by any means. If you use the code for your own projects please give  */
/* me credit. Please send a bug report. Don't alter or remove this header! */
/*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*/
#ifndef FS_SYS_FSFILEDIR_H
#define FS_SYS_FSFILEDIR_H
//---------------------------------------------------------------------------
#if defined(FS_INCLUDE_USERDEFS) // inject user definition file on request
#   include <fs/fsUserDefs.h>
#endif
//---------------------------------------------------------------------------
namespace fs { namespace sys {
//---------------------------------------------------------------------------
//! <b>F</b>ile & <b>d</b>i<b>r</b>ectory utilities.
namespace fdr {
//---------------------------------------------------------------------------
//! Returns a platform dependent path separator (as a string).
const char * GetPathSpr();
//---------------------------------------------------------------------------
//! Returns a system dependent new-line code (as a string).
const char * GetNewLn();
//---------------------------------------------------------------------------
//! Returns the app startup full directory path (with the separ. at the end).
const char * GetAppDir();
//---------------------------------------------------------------------------
//! Checks if the given full directory path exists.
bool IsDirExist(const char *pszFullDir);
//---------------------------------------------------------------------------
//! Creates a directory structure according to the given full dir. path.
bool MakeDir(const char *pszFullDir);
//---------------------------------------------------------------------------
//! Merges and decorates two given parts of a path.
void NicePath(char *pszOut, const char *pszDir, const char *pszFile = 0);
//---------------------------------------------------------------------------
//! Directory list.
/*! Listed items are stored <b>relatively</b> to the given <b>root
    directory</b> path. For obtaining the <b>full-path</b>, it is necessary
    to <b>combine</b> #GetRoot with a particular list <b>item</b>. If
    <b>directories</b> are listed as well, they are all stored <b>at the
    beginning</b> of the list (even if recursion is enabled). It is possible
    to <b>distinguish</b> between files and directories by <b>looking at the
    last character </b> of a particular item. A directory is indicated by
    the <b>path separator</b> (fs::sys::GetPathSpr).
*/
class DirLst_t
{
 public:
 
    //! Fills the list with the content of the given directory.
    explicit DirLst_t(const char *pszFullDir, bool bRecur = false,
     bool bInclDirs = false, bool bInclFiles = true);
    ~DirLst_t();
    //! Checks the internal state / correct construction.
    bool IsValid() const;
    
    //! Returns a full directory path of the root.
    const char * GetRoot() const;
    //! Returns the number of items in the list.
    int GetNum() const;
    //! Returns an item at the given position in the list.
    const char * GetAt(int nIndex) const;
     
 private:
 
    DirLst_t(); //!< No def ctor.
    DirLst_t(const DirLst_t &); //!< No copy.
    DirLst_t & operator = (const DirLst_t &); //!< No assignment.

    void *m_pImpl; //!< The PIPML idiom is used.
};
//---------------------------------------------------------------------------
}}} // namespace fs::sys:fdr
//---------------------------------------------------------------------------
#endif // FS_SYS_FSFILEDIR_H