/*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*/
/*                          _______      ______    ______          __      */
/*  ~ ~ ~ ~ ~ ~ ~ ~ ~ ~    / ____(_)___ / ___ /)  / ____/___  ____/ /__    */
/*      [fsMemMgr]        / /_  / / __ \\__ \|/  / /   / __ \/ __  / _ \   */
/*       rev. 24         / __/ / / /_/ /__/ /   / /___/ /_/ / /_/ /  __/   */
/*    20th Jan 2007     /_/   /_/ ,___/____/    \____/\____/\__,_/\___/    */
/*     [x] stable              /_/ (c) 2005-7 Filip STOKLAS (FipS)         */
/*  ~ ~ ~ ~ ~ ~ ~ ~ ~ ~       http://HOLE.4FipS.com/fips_code.php          */
/*                                                                         */
/* This code is free for personal and commercial use. You may redistribute */
/* it by any means. If you use the code for your own projects please give  */
/* me credit. Please send a bug report. Don't alter or remove this header! */
/*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*/
// NOTES:
// [o] Dependency reduction design compromises:
//     [o] 'std::allocator<void>::const_pointer' => 'const void *'
//     [o] 'std::numeric_limits<size_type>::max()' => 'size_t(-1)'
// [o] Workaround for Pocket PC 2002 env.: old SGI STL & MS eVC++ 6.0
//     [o] FS_MMGR_WORKAROUND1 defined
//     [o] nonstandard custom allocator
//     [o] 'Key_t' things...
//---------------------------------------------------------------------------
#include <fs/sys/fsMemMgr.h>

#include <fs/sys/fsMemMgr_NoT.h> // turn off the tracking macros
#include <memory>
#include <map>
#include <fs/sys/fsMemMgr.h> // turn on the tracking macros

#if defined(UNDER_CE)
#   include<windows.h> // TerminateProcess
#endif
//---------------------------------------------------------------------------
#if (_MSC_VER >= 1400) // MSVC8
#   pragma warning(disable : 4996) // function was marked as deprecated
#endif
//---------------------------------------------------------------------------
#if defined(UNDER_CE) && defined(__STL_DEFAULT_ALLOCATOR)
#   define FS_MMGR_WORKAROUND1
#endif
//---------------------------------------------------------------------------
#if defined(FS_MMGR_ENABLE_BASIC_STATS) || defined(FS_MMGR_ENABLE_TRACKING)
#   define FS_MMGR_DEF_ATEXIT
#endif
//---------------------------------------------------------------------------
// Define a 'silent' assert break with no memory allocations.
#if defined(FS_MMGR_ENABLE_TRACKING) && defined(FS_MMGR_ENABLE_RTM_BREAK)
#   if defined(_WIN32)
#       if defined(UNDER_CE)
#           define FS_MMGR_ASSERT(exp)                                      \
            {                                                               \
                bool bExp = (exp) ? true : false;                           \
                if(!bExp) TerminateProcess(GetCurrentProcess(), 0);         \
            }
#       else
#           define FS_MMGR_ASSERT(exp)                                      \
            {                                                               \
                bool bExp = (exp) ? true : false; if(!bExp) __asm int 3     \
            }
#       endif
#   else
#       define FS_MMGR_ASSERT(exp) (void(0))
#   endif
#else
#   define FS_MMGR_ASSERT(exp) (void(0))
#endif
//---------------------------------------------------------------------------
//--- ANONYMOUS NAMESPACE >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
//---------------------------------------------------------------------------
namespace {
//---------------------------------------------------------------------------
const char *GetModNm();
//---------------------------------------------------------------------------
//! The name of the static module (mainly for file output).
char g_szModNm[32] = {'\0'};
//---------------------------------------------------------------------------
#if defined(FS_MMGR_DEF_ATEXIT)
//! A flag indicating that '::atexit' has already been set.
bool g_bAtExitSet = false;
#endif
//---------------------------------------------------------------------------
#if defined(FS_MMGR_ENABLE_BASIC_STATS)
//! Basic statistics.
class Stats_t
{
 public:
 
    Stats_t() : m_nAlloc(0), m_nFree(0), m_nMax(0) {}
     
    void Alloc()
    {
        ++m_nAlloc;
        
        // update the maximum number of allocations
        const int nDelta = m_nAlloc - m_nFree;
        if(nDelta > m_nMax) m_nMax = nDelta;
    }
    
    void Free() { ++m_nFree; }
    
    int GetNumLeaks() const { return m_nAlloc - m_nFree; }
    
    void Dump(FILE *hOut) const
    {
        int nLeaks = GetNumLeaks();
    
        ::fprintf(hOut, "+--------------------------------------------+\n");
        ::fprintf(hOut, "| FipS' MEMORY MANAGER (basic stats)         |\n");
        ::fprintf(hOut, "| module: %-34s |\n", GetModNm());
        ::fprintf(hOut, "+--------------------------------+-----------+\n");
        ::fprintf(hOut, "|    Total number of allocations | %-9d |\n",
         m_nAlloc);
        ::fprintf(hOut, "|  Total number of deallocations | %-9d |\n",
         m_nFree);
        ::fprintf(hOut, "|      Max number of allocations | %-9d |\n",
         m_nMax);
        ::fprintf(hOut, "|                Number of leaks | %-7d %c |\n",
         nLeaks, nLeaks > 0 ? '!' : ' ' );
        ::fprintf(hOut, "+--------------------------------+-----------+\n");
    }
    
 private:
 
    int m_nAlloc;
    int m_nFree;
    int m_nMax;
    
} g_Stats;
#endif
//---------------------------------------------------------------------------
#if defined(FS_MMGR_ENABLE_TRACKING)
//! Memory operation types.
enum MemOp_e
{
    MOP_NEW,
    MOP_DELETE,
    MOP_NEW_ARRAY,
    MOP_DELETE_ARRAY
};
#endif
//---------------------------------------------------------------------------
#if defined(FS_MMGR_ENABLE_TRACKING)
//! Memory operation names (keep in sync with MemOp_e).
const char * const apszMemOpNms[] =
{
    "new",
    "delete",
    "new []",
    "delete []"
};
#endif
//---------------------------------------------------------------------------
#if defined(FS_MMGR_ENABLE_TRACKING)
//! Memory operation pairs.
const MemOp_e aMemOpPrs[][2] = 
{
    { MOP_NEW,       MOP_DELETE },
    { MOP_NEW_ARRAY, MOP_DELETE_ARRAY }
};
#endif
//---------------------------------------------------------------------------
#if defined(FS_MMGR_ENABLE_TRACKING)
//! Copies and crops a string.
void StrCpyCrp(char *pszDest, const char *pszSrc, size_t uMaxLen)
{
    size_t uLen = ::strlen(pszSrc);
            
    if(uLen > uMaxLen)
    {
        size_t uCount = uMaxLen - 2;
        pszDest[0] = pszDest[1] = pszDest[2] = '.';
        ::memcpy(pszDest + 3, pszSrc + uLen - uCount + 1, uCount);
    }
    else
        ::strcpy(pszDest, pszSrc);
    
}
#endif
//---------------------------------------------------------------------------
#if defined(FS_MMGR_ENABLE_TRACKING)
//! Code position holder.
class CodePos_t
{
 public:

    char m_szFunc[32];
    char m_szFile[32];
    int  m_nLine;
        
    CodePos_t() { m_szFunc[0] = m_szFile[0] = '?'; m_szFunc[1] =
     m_szFile[1] = '\0'; m_nLine = -1; }
     
    CodePos_t(const char *pszFunc, const char *pszFile, int nLine)
    {
        *this = CodePos_t(); // reset the state
        
        if(pszFunc) StrCpyCrp(m_szFunc, pszFunc, sizeof(m_szFunc) - 1);
        if(pszFile) StrCpyCrp(m_szFile, pszFile, sizeof(m_szFile) - 1);
        
        m_nLine = nLine;
    }
    
} g_LastPos;
#endif
//---------------------------------------------------------------------------
#if defined(FS_MMGR_ENABLE_TRACKING)
#pragma warning(push)
#pragma warning(disable: 4100) // unreferenced formal parameter
//---
// It is necessary to provide a custom STL memory allocator in order to
// prevent recursion during internal (de)allocations.
//---
#if defined(FS_MMGR_WORKAROUND1)
//---
template<class T>
class Alloc_T // the simplified SGI's allocator interface
{
 public:

    // call the global malloc / free
    static T* allocate(size_t n)
     { return reinterpret_cast<T*>(::malloc(n * sizeof(T))); }
    static T* allocate(void)
     { return reinterpret_cast<T*>(::malloc(sizeof(T))); }
    static void deallocate(void *p, size_t n)
     { ::free(p); }
    static void deallocate(void *p)
     { ::free(p); }
};
//---
#else
//---
template <class T>
class Alloc_T // the standard allocator interface
{
 public:
 
    typedef T value_type;
    typedef T* pointer;
    typedef const T* const_pointer;
    typedef T& reference;
    typedef const T& const_reference;
    typedef size_t size_type;
    typedef ptrdiff_t difference_type;
    
    template <class U>
    struct rebind { typedef Alloc_T<U> other; };
    
    Alloc_T() {}
    Alloc_T(const Alloc_T &) {}
#if !defined(_MSC_VER) || _MSC_VER >= 1300 //!< \todo check exact VER
    template <class U>
    Alloc_T(const Alloc_T<U> &) {}
#endif
    ~Alloc_T() {}
        
    pointer address(reference r) const { return &r; }
    const_pointer address(const_reference r) const { return &r; }
            
    // call the global malloc / free
    pointer allocate(size_type n, const void * = 0)
     { return reinterpret_cast<pointer>(::malloc(n * sizeof(T))); }
    void deallocate(pointer p, size_type) { ::free(p); }
            
    size_type max_size() const { return size_t(-1) / sizeof(T); }
            
    void construct(pointer p, const T &t) { new(p) T(t); }
    void destroy(pointer p) { p->~T(); }
};
//---
template <class T, class U>
bool operator == (const Alloc_T<T> &, const Alloc_T<U> &)
{ return true; }
//---
template <class T, class U>
bool operator != (const Alloc_T<T> &LHS, const Alloc_T<U> &RHS)
{ return false; }
//---
template <>
class Alloc_T<void>
{
 public:
    
    typedef void value_type;
    typedef void* pointer;
    typedef const void* const_pointer;
    typedef size_t size_type;
    typedef ptrdiff_t difference_type;
    
    template <class U>
    struct rebind { typedef Alloc_T<U> other; };
};
//---
#endif
//---
#pragma warning(pop)
#endif
//---------------------------------------------------------------------------
#if defined(FS_MMGR_ENABLE_TRACKING)
//! Memory tracker.
class Tracker_t
{
 public:
 
    Tracker_t() : m_nCurMem(0), m_nMaxMem(0) {}
 
    void Alloc(const void *pAddr, int nSize, MemOp_e eMemOp,
     const CodePos_t &Pos)
    {
        FS_MMGR_ASSERT(pAddr);

        Val_t Val(nSize, eMemOp, Pos);
        m_Map.insert(Pair_t(pAddr, Val));
        m_nCurMem += nSize;
        
        // update the maximum amount of used memory
        if(m_nCurMem > m_nMaxMem) m_nMaxMem = m_nCurMem;
        
        g_LastPos = CodePos_t(); // reset the last code position
    }

    void Free(const void *pAddr, MemOp_e eMemOp)
    {
        if(!pAddr) // according to the standard it is correct
            return;
            
        Map_t::iterator it = m_Map.find(pAddr);

        if(it == m_Map.end())
        {
            //     RUNTIME ERROR: An unknown / bad memory address 'pAddr'
            //                    has been passed!
            FS_MMGR_ASSERT(0); // ^^^
            return;
        }

        MemOp_e AllocMemOp = (*it).second.m_MemOp;
        bool bFound = false;
        for(int i = 0; i < sizeof(aMemOpPrs) / sizeof(aMemOpPrs[0]); ++i)
            if(
             AllocMemOp == aMemOpPrs[i][0] &&
             eMemOp == aMemOpPrs[i][1]
            )
                { bFound = true; break; }
            
        if(!bFound)
        {
            //     RUNTIME ERROR: Wrong deallocation method detected!
            //                    (see 'AllocMemOp' vs. 'FreeMemOp')
            FS_MMGR_ASSERT(0); // ^^^
            return;
        }
        
        m_nCurMem -= (*it).second.m_nSize;
        m_Map.erase(it);
    }
    
    int GetNumLeaks() const { return static_cast<int>(m_Map.size()); }
    
    void Dump(FILE *hOut) const
    {
        struct Local_t
        {
            static const char * DecorSize(int nSize)
            {
                static char szOut[32];
                                
                if(nSize >= 1E6)
                    ::sprintf(szOut, "%.2f MB", nSize / (float)1E6);
                else if(nSize >= 1E3)
                    ::sprintf(szOut, "%.2f KB", nSize / (float)1E3);
                else
                    ::sprintf(szOut, "%d B", nSize);
                    
                return szOut;
            }
        };
        
        bool bEmpty = m_Map.empty();
    
        ::fprintf(hOut, "\n+------------------------------------------+\n");
        ::fprintf(hOut, "| FipS' MEMORY MANAGER (tracking details)  |\n");
        ::fprintf(hOut, "| module: %-32s |\n", GetModNm());
        ::fprintf(hOut, "+------------------------------+-----------+\n");
        ::fprintf(hOut, "|    Max allocated memory size | %-9s |\n",
         Local_t::DecorSize(m_nMaxMem));
        ::fprintf(hOut, "|              Number of leaks | %-7d %c |\n",
         m_Map.size(), bEmpty ? ' ' : '!' );
        ::fprintf(hOut, "+------------------------------+--%c--------+",
         bEmpty ? '-' : '+');
        
        if(bEmpty) { ::fprintf(hOut, "\n"); return; }
        
        ::fprintf(hOut, "---------------------------------"
         "+------------+-----------+\n");
         
        for(Map_t::const_iterator it = m_Map.begin();
         it != m_Map.end(); ++it)
        {
            const CodePos_t &Pos = (*it).second.m_Pos;
            ::fprintf(hOut, "| %31s | %-6d | %-31s | 0x%08x | %-9s |\n",
             Pos.m_szFile, Pos.m_nLine, Pos.m_szFunc, (size_t)
             ((*it).first.m_pAddr), Local_t::DecorSize((*it).second.m_nSize)
            );
        }
        
        ::fprintf(hOut, "+---------------------------------+--------+"
         "---------------------------------+------------+-----------+\n");
    }
 
 private:
 
    struct KeyT_t
    {
        KeyT_t(const void *pAddr) : m_pAddr(pAddr) {}
        bool operator < (const KeyT_t &RHS) const
         { return m_pAddr < RHS.m_pAddr; }
         
        const void *m_pAddr;
                
     private:
     
        KeyT_t();
    };
 
    struct ValT_t
    {
        int m_nSize;
        MemOp_e m_MemOp;
        CodePos_t m_Pos;
        
        ValT_t(int nSize, MemOp_e MemOp, CodePos_t Pos):
         m_nSize(nSize), m_MemOp(MemOp), m_Pos(Pos) {}
         
     private:
     
        ValT_t();
    };

#if defined(FS_MMGR_WORKAROUND1)
    typedef const KeyT_t Key_t;
    typedef const ValT_t Val_t;
#else
    typedef KeyT_t Key_t;
    typedef ValT_t Val_t;
#endif

    typedef std::pair<Key_t, Val_t> Pair_t;
    typedef std::map<Key_t, Val_t, std::less<Key_t>, Alloc_T<Val_t> > Map_t;
     
    Map_t m_Map;
    
    int m_nCurMem;
    int m_nMaxMem;
    
} g_Tracker;
#endif
//---------------------------------------------------------------------------
#if defined(FS_MMGR_DEF_ATEXIT)
void AtExit()
{
    char pszFile[256];
    
#if defined(UNDER_CE)

    // there is no concept of the working directory so it is necessary
    // to extract one from the module filename
     
    wchar_t wszTemp[MAX_PATH + 1];
    ::GetModuleFileName(0, wszTemp, MAX_PATH);
    
    size_t i = 0; while(i < MAX_PATH) { char ch = static_cast<wchar_t>
     (wszTemp[i]); pszFile[i] = ch; if(ch == '\0') break; ++i; }
     
    char *pszSep = ::strrchr(pszFile, '\\');
    FS_MMGR_ASSERT(pszSep);
    *(pszSep + 1) = '\0';
    
    ::strcat(pszFile, "fs_mmgr_");
    ::strcat(pszFile, GetModNm());
    ::strcat(pszFile, ".log");
    
#else

    ::sprintf(pszFile, "fs_mmgr_%s.log", GetModNm());

#endif

    FILE *hOutFl = ::fopen(pszFile, "w");
            
    if(!hOutFl)
        return;
        
    bool bLeaks = false;
    
#if defined(FS_MMGR_ENABLE_BASIC_STATS)
     g_Stats.Dump(hOutFl);
     bLeaks |= g_Stats.GetNumLeaks() > 0 ? true : false;
#endif

#if defined(FS_MMGR_ENABLE_TRACKING)
     g_Tracker.Dump(hOutFl);
     bLeaks |= g_Tracker.GetNumLeaks() > 0 ? true : false;
#endif
    
    ::fclose(hOutFl);
    
    if(bLeaks)
    {
        //     RUNTIME ERROR: Memory leaks detected, for more details see
        //                    the log file (if the tracking is enabled).
        FS_MMGR_ASSERT(0); // ^^^
    }
}
#endif
//---------------------------------------------------------------------------
#if defined(FS_MMGR_OWN_OPERATORS)
void * Alloc(size_t uSize)
{
#if defined(FS_MMGR_DEF_ATEXIT)
    if(!g_bAtExitSet)
    {
        ::atexit(AtExit);
        g_bAtExitSet = true;
    }
#endif

#if defined(FS_MMGR_ENABLE_BASIC_STATS)
    g_Stats.Alloc();
#endif

#if defined(FS_MMGR_ENABLE_POOL)
#   error Not implemented!
    return 0;
#else
    return ::malloc(uSize);
#endif
}
#endif
//---------------------------------------------------------------------------
#if defined(FS_MMGR_OWN_OPERATORS)
void Free(void *pAddr)
{
#if defined(FS_MMGR_ENABLE_BASIC_STATS)
    g_Stats.Free();
#endif

#if defined(FS_MMGR_ENABLE_POOL)
#   error Not implemented!
#else
    ::free(pAddr);
#endif
}
#endif
//---------------------------------------------------------------------------
/*! The default module name is formed using the address of the function
    itself, which is unique in each static module...
*/
const char * GetModNm()
{
    if(g_szModNm[0] == '\0')
        ::sprintf(g_szModNm, "0x%08x", (size_t)&GetModNm);
    
    return g_szModNm;
}
//---------------------------------------------------------------------------
} // anonymous namespace
//---------------------------------------------------------------------------
//--- NAMESPACE fs::sys::mem >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
//---------------------------------------------------------------------------
/*! The function sets a name of the current static module, this affects
    the name of the resulting log file. It is useful when e.g. there are
    many DLL modules involved and the memory tracking is performed within
    each of them, then it is possible to distinguish between resulting log
    files. By default the name is formed using a static memory address from
    the underlaying module so that there is always a log file for each of
    the modules.
*/
void fs::sys::mmgr::SetModNm(const char *pszModNm)
{
    if(pszModNm)
    {
        ::strncpy(g_szModNm, pszModNm, sizeof(g_szModNm));
        g_szModNm[sizeof(g_szModNm) - 1] = '\0';
    }
    else
        g_szModNm[0] = '\0';
}
//---------------------------------------------------------------------------
#if defined(FS_MMGR_ENABLE_TRACKING)
void fs::sys::mmgr::aux::Track(const char *pszFunc, const char *pszFile,
 int nLine)
{
    g_LastPos = CodePos_t(pszFunc, pszFile, nLine);
}
#endif
//---------------------------------------------------------------------------
//--- MEMORY ALLOCATION OPERATORS >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
//---------------------------------------------------------------------------
#include <fs/sys/fsMemMgr_NoT.h> // turn off the tracking macros
//---------------------------------------------------------------------------
#if defined(FS_MMGR_OWN_OPERATORS)
void * operator new (size_t uSize)
{
    void *pAddr = Alloc(uSize);
    
#if defined(FS_MMGR_ENABLE_TRACKING)
    g_Tracker.Alloc
     (pAddr, static_cast<int>(uSize), MOP_NEW, g_LastPos);
#endif
    
    return pAddr;
}
#endif
//---------------------------------------------------------------------------
#if defined(FS_MMGR_OWN_OPERATORS)
void operator delete (void *pAddr)
{
#if defined(FS_MMGR_ENABLE_TRACKING)
    g_Tracker.Free(pAddr, MOP_DELETE);
#endif

    Free(pAddr);
}
#endif
//---------------------------------------------------------------------------
#if defined(FS_MMGR_OWN_OPERATORS)
void * operator new [] (size_t uSize)
{
    void *pAddr = Alloc(uSize);
    
#if defined(FS_MMGR_ENABLE_TRACKING)
    g_Tracker.Alloc
     (pAddr, static_cast<int>(uSize), MOP_NEW_ARRAY, g_LastPos);
#endif
    
    return pAddr;
}
#endif
//---------------------------------------------------------------------------
#if defined(FS_MMGR_OWN_OPERATORS)
void operator delete [] (void *pAddr)
{
#if defined(FS_MMGR_ENABLE_TRACKING)
    g_Tracker.Free(pAddr, MOP_DELETE_ARRAY);
#endif

    Free(pAddr);
}
#endif
//---------------------------------------------------------------------------
#include <fs/sys/fsMemMgr.h> // turn on the tracking macros
//---------------------------------------------------------------------------