/*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*/
/*                          _______      ______    ______          __      */
/*  ~ ~ ~ ~ ~ ~ ~ ~ ~ ~    / ____(_)___ / ___ /)  / ____/___  ____/ /__    */
/*      [fsMemMgr]        / /_  / / __ \\__ \|/  / /   / __ \/ __  / _ \   */
/*       rev. 24         / __/ / / /_/ /__/ /   / /___/ /_/ / /_/ /  __/   */
/*    20th Jan 2007     /_/   /_/ ,___/____/    \____/\____/\__,_/\___/    */
/*     [x] stable              /_/ (c) 2005-7 Filip STOKLAS (FipS)         */
/*  ~ ~ ~ ~ ~ ~ ~ ~ ~ ~       http://HOLE.4FipS.com/fips_code.php          */
/*                                                                         */
/* This code is free for personal and commercial use. You may redistribute */
/* it by any means. If you use the code for your own projects please give  */
/* me credit. Please send a bug report. Don't alter or remove this header! */
/*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*/
/*! \file fsMemMgr_NoT.h
    \brief Memory manager wrapping macros disabler.
    
    In some cases, it is important to have a chance to <b>disable</b> the
    '<b>new</b>' and '<b>delete</b>' macro wrappers locally, which is the
    main purpose of this header. Since the <b>macro expansion make sense</b>
    only in the case of memory <b>(de)allocation</b>. In other cases (e.g.
    when <b>operator overloading</b> takes place in code) the macros <b>
    result in syntactically wrong statements</b>. Note that if the <b>macro
    wrappers are disabled</b> the memory <b>tracking</b> itself <b>is still
    working</b> (through overloaded 'new' and 'delete' operators), only
    detailed code position information won't be available. See the example
    of usage below.
       
    \code
    #include <fs/sys/fsMemMgr_NoT.h> // turn off the tracking macros
    #include <vector>
    #include <fs/sys/fsMemMgr.h> // turn on the tracking macros
    \endcode
*/
//---------------------------------------------------------------------------
#if defined(FS_INCLUDE_USERDEFS) // inject user definition file on request
#   include <fs/fsUserDefs.h>
#endif
//---------------------------------------------------------------------------
//--- EXTERNAL CONFIGURATION >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
//---------------------------------------------------------------------------
#if !defined(FS_MMGR_USERDEFS) // define to overwrite default definitions
#   if !defined(NDEBUG)
        //! Enable: Basic statistics.
        /*! If this directive is defined, only basic statistics are
            collected and written into the log file. This method has
            almost zero overhead... */
#       define FS_MMGR_ENABLE_BASIC_STATS
        //! Enable: Full memory tracking.
        /*! If this directive is defined, the full memory tracking is
            enabled and results are written into the log file. */
#       define FS_MMGR_ENABLE_TRACKING
        //! Enable: Runtime break invoking in case of memory error.
        /*! After defining the directive, the memory manager invokes a
            runtime error at any time a memory error occurs. It helps
            catch memory errors in runtime as soon as possible. */
#       define FS_MMGR_ENABLE_RTM_BREAK
#   endif // NDEBUG
#endif // FS_MMGR_USERDEFS
//---------------------------------------------------------------------------
//--- UNDEFINE TRACKING MACROS >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
//---------------------------------------------------------------------------
#if defined(FS_MMGR_ENABLE_TRACKING)
//---------------------------------------------------------------------------
// get rid of the tracking macro
#if defined (new)
#undef new
#endif
//---------------------------------------------------------------------------
// get rid of the tracking macro
#if defined (delete)
#undef delete
#endif
//---------------------------------------------------------------------------
#endif // FS_MMGR_ENABLE_TRACKING
//---------------------------------------------------------------------------
//--- <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
//---------------------------------------------------------------------------