/*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*/
/*                          _______      ______    ______          __      */
/*  ~ ~ ~ ~ ~ ~ ~ ~ ~ ~    / ____(_)___ / ___ /)  / ____/___  ____/ /__    */
/*     [fsXmlDoc]         / /_  / / __ \\__ \|/  / /   / __ \/ __  / _ \   */
/*       rev. 4          / __/ / / /_/ /__/ /   / /___/ /_/ / /_/ /  __/   */
/*    13th Jan 2007     /_/   /_/ ,___/____/    \____/\____/\__,_/\___/    */
/*     [x] stable              /_/ (c) 2005-7 Filip STOKLAS (FipS)         */
/*  ~ ~ ~ ~ ~ ~ ~ ~ ~ ~       http://HOLE.4FipS.com/fips_code.php          */
/*                                                                         */
/* This code is free for personal and commercial use. You may redistribute */
/* it by any means. If you use the code for your own projects please give  */
/* me credit. Please send a bug report. Don't alter or remove this header! */
/*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*/
#ifndef FS_UT_FSXMLDOC_H
#define FS_UT_FSXMLDOC_H
//---------------------------------------------------------------------------
#if defined(FS_INCLUDE_USERDEFS) // inject user definition file on request
#   include <fs/fsUserDefs.h>
#endif
//---------------------------------------------------------------------------
namespace fs { namespace ut {
//---------------------------------------------------------------------------
//! Simple TinyXml wrapper.
/*! The class wraps <a href="http://www.grinninglizard.com/tinyxml/">
    TinyXml</a>. Furthermore XmlDoc_t offers very simple and powerful
    <b>string-based query</b> requests for <b>obtaining data</b> from the
    underlying <b>XML</b> document. A <b>query</b> is formed according to
    the following <b>rules</b>: '<code>^</code>' defines the active
    traversation <b>branch</b>, '<code>#</code>' <b>counts children</b>
    elements with the given name, '<code>\@</code>' obtains a <b>value of
    element's attribute</b>. For example
    '<code>disk/folder^2/file^5/\@size</code>' obtains a value of the
    <i>size</i> attribute of the 6-th <i>file</i> in the 3-rd
    <i>folder</i> on the <i>disk</i>. For more details see the example.
*/
class XmlDoc_t
{
 public:
 
    //! Loads the XML document using the given filename.
    explicit XmlDoc_t(const char *pszFile);
    //! Parses the XML document using the given string data.
    XmlDoc_t(const char *pszData, bool bUnused);
    //! Unloads the XML document.
    ~XmlDoc_t();
    //! Checks the internal state / correct construction.
    bool IsValid() const;
    
    //! Queries the XML document, obtains an int result.
    bool Query(const char *pszQuery, int &nValue) const;
    //! Queries the XML document, obtains a string result.
    bool Query(const char *pszQuery, const char *&pszText) const;
     
 private:
 
    XmlDoc_t(); //!< No def ctor.
    XmlDoc_t(const XmlDoc_t &); //!< No copy.
    XmlDoc_t & operator = (const XmlDoc_t &); //!< No assignment.

    void *m_pImpl; //!< The PIPML idiom is used.
};
//---------------------------------------------------------------------------
}} // namespace fs::ut
//---------------------------------------------------------------------------
#endif // FS_UT_FSXMLDOC_H